import pytest
import os,sys,inspect
import numpy as np
import pyscal3.core as pc


def test_rdf_bcc():
    sys = pc.System.create.lattice.bcc(repetitions = [10, 10, 10])
    rdf, r = sys.calculate.radial_distribution_function(rmax=2)

    args = np.argsort(rdf)[::-1]
    assert(r[args[0]]-0.86 < 1E-5)

def test_rdf_fcc():
    sys = pc.System.create.lattice.fcc(repetitions = [10, 10, 10])
    rdf, r = sys.calculate.radial_distribution_function(rmax=2)

    args = np.argsort(rdf)[::-1]
    assert(r[args[0]]-0.70 < 1E-5)



import numpy as np

def calculate_sro(system, reference_type=1, compare_type=2, average=True):
    """
    Calculate short range order

    Parameters
    ----------
    reference_type: int, optional
        type of the atom to be used a reference. default 1

    compare_type: int, optional
        type of the atom to be compared to. default 2

    average: bool, optional
        if True, average over all atoms of the reference type in the system.
        default True.

    Returns
    -------
    vec: list of float
        The short range order averaged over the whole system for atom of
        the reference type. Only returned if `average` is True. 

    Notes
    -----
    Calculates the short range order for an AB alloy using the approach by
    Cowley [1]. Short range order is calculated as,

    .. math::

        \\alpha_i = 1 - \\frac{n_i}{m_A c_i}

    where n_i is the number of atoms of the non reference type among the c_i atoms
    in the ith shell. m_A is the concentration of the non reference atom. Please
    note that the value is calculated for shells 1 and 2 by default. In order for
    this to be possible, neighbors have to be found first using the :func:`~pyscal.core.System.find_neighbors`
    method. The selected neighbor method should include the second shell as well. For this
    purpose `method=cutoff` can be chosen with a cutoff long enough to include the second
    shell. In order to estimate this cutoff, one can use the :func:`~pyscal.core.System.calculate_rdf`
    method.

    The method also works for alloying elements greater than 2 [2]. For this, you can choose the reference type.

    References
    ----------
    .. [1] Cowley J. M., PR 77(5), 1950.
    .. [2] de Fountaine D., J. Appl. Cryst. 4(15), 1971.

    """

    def _get_dict(val, cdict):
        cx, cxc = np.unique(val, return_counts=True)
        cx = list(cx)
        cxc = list(cxc)
        for key, val in cdict.items():
            if key not in cx:
                cx.append(key)
                cxc.append(0)
        d = dict([(x, cxc[c]/np.sum(cxc)) for c, x in enumerate(cx)])
        return d

    system._check_neighbors()
    
    cdict = system.atoms.composition_ints

    neighbortypes = system.atoms["types"][system.atoms['neighbors']]
    comp_dicts = [_get_dict(neighbortype, cdict) for neighbortype in neighbortypes]
    local_comp = np.array([d[reference_type] for d in comp_dicts])
    global_comp = cdict[reference_type]

    if reference_type == compare_type:
        sro = (local_comp - global_comp)/(1 - global_comp)
    else:
        sro = 1 - (local_comp/global_comp)

    system.atoms["sro"]= sro
    mapdict = {}
    mapdict['chemical'] = {}
    mapdict['chemical']['short_range_order'] = "sro"
    system.atoms._add_attribute(mapdict)

    if average:
        np.mean(sro)
        


import pyscal3.core as pc
import os
import numpy as np
from ase.build import bulk
from pyscal3.atoms import Atoms


def test_system_init():
	sys = pc.System.create.lattice.bcc(repetitions = [10,10,10], lattice_constant=3.127)
	assert len(sys.atoms["positions"]) == 10*10*10*2
	assert sys.triclinic == 0
	assert np.abs(sys.boxdims[0] - sys.box[0][0]) < 1E-5


def test_system_triclinic():
	struct = bulk('Cu').repeat(10)
	sys = pc.System()
	sys.box = np.array(struct.cell)
	atoms = {}
	atoms["positions"] = struct.positions
	sys.atoms = atoms
	assert sys.triclinic == 1
	tb = (sys.rot== np.array(struct.cell).T)
	assert np.prod(tb) == 1
	tb = (sys.rotinv==np.linalg.inv(np.array(struct.cell).T))
	assert np.prod(tb) == 1

def test_nop():
	sys = pc.System.create.lattice.bcc(repetitions = [2, 2, 2], lattice_constant=3.127)

	assert sys.natoms == 16
	assert len(sys.atoms['positions']) == 128

	for a in sys.iter_atoms():
		assert np.sum(a["positions"]) == 0
		break

	natoms = {'positions':[[0,0,0]]}
	sys.atoms += natoms
	assert sys.natoms == 17

def test_embed():
	cu = bulk('Cu')
	sys = pc.System()
	sys.box = np.array(cu.cell)
	sys.atoms = Atoms({"positions": cu.positions})
	sys.modify.embed_in_cubic_box()
	assert np.abs(sys.box[0][0] - 5.105310960166873) < 1E-5

def test_distance():
	sys = pc.System.create.lattice.bcc(repetitions = [2, 2, 2], lattice_constant=3.127)
	dist = sys.calculate.distance([0.0, 0.0, 0.0], [1.5635, 1.5635, 1.5635])
	assert np.abs(dist - 2.708061437633939) < 1E-5	

def test_composition():
	sys = pc.System.create.lattice.l12(repetitions = [2, 2, 2], lattice_constant=3.127)
	c = sys.concentration
	assert c[1] == 0.25
	assert c[2] == 0.75

	c = sys.composition
	assert c[1] == 0.25
	assert c[2] == 0.75

def test_volume():
	sys = pc.System.create.lattice.fcc(repetitions = [10, 10, 10])
	assert sys.volume == 1000

def test_system_init():
	sys = pc.System.create.lattice.custom([[0, 0, 0], [0.5, 0.5, 0.5]], [1, 2], 
		[[1,0,0], [0,1,0], [0,0,1]])
	assert sys.natoms == 2


import pytest
import os,sys,inspect
import numpy as np
import pyscal3.core as pc
from ase.build import bulk

#test conventional cna for fcc
def test_cna_cutoff():
    sys = pc.System.create.lattice.fcc(repetitions=(7,7,7), lattice_constant=4.00)
    res = sys.calculate.common_neighbor_analysis(lattice_constant=4.00)
    assert res["fcc"] == 7*7*7*4

def test_cna_a1():
    sys = pc.System.create.lattice.fcc(repetitions=(7,7,7), lattice_constant=4.00)
    res = sys.calculate.common_neighbor_analysis(lattice_constant=None)
    assert res["fcc"] == 7*7*7*4

#now test adaptive
def test_cna_adaptive():
    sys = pc.System.create.lattice.fcc(repetitions=(7,7,7), lattice_constant=4.00)
    sys.calculate.common_neighbor_analysis(lattice_constant=None)
    assert sys.atoms.structure[0] == 1

    sys = pc.System.create.lattice.hcp(repetitions=(7,7,7), lattice_constant=4.00)
    sys.calculate.common_neighbor_analysis(lattice_constant=None)
    assert sys.atoms.structure[0] == 2

    sys = pc.System.create.lattice.bcc(repetitions=(7,7,7), lattice_constant=4.00)
    sys.calculate.common_neighbor_analysis(lattice_constant=None)
    assert sys.atoms.structure[0] == 3


def test_ase_bulks():
    
    al_fcc = bulk("Al")
    fe_bcc = bulk("Fe")
    ti_hcp = bulk("Ti")

    sys = pc.System()
    sys.read.ase(al_fcc)
    cna = sys.calculate.common_neighbor_analysis()
    assert cna["fcc"] == 1

    sys = pc.System()
    sys.read.ase(fe_bcc)
    cna = sys.calculate.common_neighbor_analysis()
    assert cna["bcc"] == 1

    sys = pc.System()
    sys.read.ase(ti_hcp)
    cna = sys.calculate.common_neighbor_analysis()
    assert cna["hcp"] == 2

def test_cna_diamond():
    sys = pc.System.create.lattice.diamond(repetitions=(7,7,7), lattice_constant=4.00)
    sys.calculate.diamond_structure()
    assert sys.atoms.structure[0] == 5


import pytest
import os
import numpy as np
import pyscal3.core as pc
import pyscal3.structure_creator as pcs
import pyscal3.traj_process as ptp

def test_create_multislice_dump():
    """
    Create a multitest dump file and test it
    """
    sys = pc.System.create.lattice.bcc(repetitions = [6, 6, 6])
    ptp.write_file(sys, "tests/bcc1.dump")

    atoms2, boxdims2 = pcs.make_crystal('bcc', repetitions=[6,6,6])
    #modify the coordinates of one atom
    x  = atoms2["positions"][0]
    x[0] += 0.01
    atoms2["positions"][0] = x
    assert len(atoms2["positions"]) == 432
    #write it out
    sys2 = pc.System()
    sys2.box = boxdims2
    sys2.atoms = atoms2

    ptp.write_file(sys2, "tests/bcc2.dump")

    #now cleanup
    if os.path.exists("tests/bcc1.dat"):
        os.remove("tests/bcc1.dat")
    if os.path.exists("tests/bcc2.dat"):
        os.remove("tests/bcc2.dat")

def test_customvals_dump():
    """
    Test writing customvals
    """
    sys = pc.System.create.lattice.bcc(repetitions = [1, 1, 1])


    #test for multiple customvals
    customks = ['one']
    customvs = [[1],[1]]
    ptp.write_file(sys, "tests/bcc4.dump", customkeys=customks, customvals=customvs)

    #now read this file
    lines = []
    for line in open("tests/bcc4.dump", 'r'):
        lines.append(line)

    #now check the atoms
    last1line = lines[-1].strip().split()
    last2line = lines[-2].strip().split()
    last3line = lines[-3].strip().split()

    #now verify
    assert last1line[-1] == '1'
    assert last2line[-1] == '1'
    assert last3line[-1] == 'one'

    #clean up
    if os.path.exists("tests/bcc4.dat"):
        os.remove("tests/bcc4.dat")

    #test for multiple customvals
    customks = ['one', 'two']
    customvs = [[1,1], [2,2]]
    ptp.write_file(sys, "tests/bcc4.dump", customkeys=customks, customvals=customvs)

    #now read this file
    lines = []
    for line in open("tests/bcc4.dump", 'r'):
        lines.append(line)

    #now check the atoms
    last1line = lines[-1].strip().split()
    last2line = lines[-2].strip().split()
    last3line = lines[-3].strip().split()

    #now verify
    assert last1line[-1] == '2'
    assert last1line[-2] == '2'
    assert last2line[-1] == '1'
    assert last2line[-2] == '1'
    assert last3line[-1] == 'two'
    assert last3line[-2] == 'one'

    #clean up
    if os.path.exists("tests/bcc4.dat"):
        os.remove("tests/bcc4.dat")



import numpy as np
import copy
from scipy.spatial import distance
import pyscal3.csystem as pc

#from pyscal3.attributes import DocumentedKeywords
def pad_repeat(atoms, box, repetitions, ghost=True):
    """
    This method creates a padded layer around the atoms
    """
    box = np.array(copy.copy(box))        
    idstart = len(atoms) + 1

    pos = atoms.positions
    nop = len(pos)
    ids = atoms.ids
    head = [x for x in range(len(pos))]
    ghosts = [False for x in range(nop)]


    #all the other keys
    datadict = {key: atoms[key][:atoms.nreal] for key in atoms.keys()}
    del datadict['positions']
    del datadict['ids']
    del datadict['head']
    del datadict['ghost']

    #prepopulate
    for d in range(3):
        #first do to the right
        pos_list = []
        new_id_list = []
        ghost_list = []

        for i in range(1, repetitions[d]):
            npos = copy.copy(pos)
            npos = npos + i*box[d]
            pos_list.append(npos)
            new_ids = [idstart+i for i in range(len(pos))]
            new_id_list.append(new_ids)
            ghost_list.append([ghost for x in range(len(pos))])
            idstart = idstart + len(new_ids)

        for i in range(-repetitions[d]+1, 0):
            npos = copy.copy(pos)
            npos = npos + i*box[d]
            pos_list.append(npos)
            new_ids = [idstart+i for i in range(len(pos))]
            new_id_list.append(new_ids)
            ghost_list.append([ghost for x in range(len(pos))])
            #change id start
            idstart = idstart + len(new_ids)

        pos = np.concatenate((pos, *pos_list))
        ids = np.concatenate((ids, *new_id_list))
        ghosts = np.concatenate((ghosts, *ghost_list))
        
        head = np.concatenate((head, np.tile(head, len(pos_list))))

        for key in datadict.keys():
            datadict[key] = np.concatenate((datadict[key], np.tile(datadict[key], len(pos_list))))

    atoms["positions"] = pos
    atoms["ids"] = ids
    atoms["head"] = head
    atoms["ghost"] = ghosts

    for key in datadict.keys():
        atoms[key] = datadict[key]


    box[0] = (2*repetitions[0]-1)*np.array(box[0])
    box[1] = (2*repetitions[1]-1)*np.array(box[1])
    box[2] = (2*repetitions[2]-1)*np.array(box[2])

    return atoms, box


def repeat(system, repetitions, 
    ghost = False, scale_box = True, 
    atoms = None, return_atoms = False):
    """
    Repeat the given system

    Parameters
    ----------
    system: pyscal System object
        the input system to be repeated

    repetitions: tuple of ints
        number of times the system is to be rotated in each direction

    ghost: bool, optional
        if True, make the new atoms ghost, default False

    scale_box: bool, optional
        if True, scale the simulation box, default True

    atoms: None, optional
        if provided use the given atoms, and not the atoms from the system

    return_atoms: bool, optional
        if True, return atoms instead of adding them to the system. Default False
    
    Returns
    -------
    system: pyscal System object
        the system with repetitions. Only returned if `return_atoms` is False.
    
    atoms: Atoms object
        only returned if `return_atoms` is True.


    """
    box = np.array(copy.copy(system.box))        
    system.actual_box = box.copy()

    if atoms is None:
        atoms = system.atoms

    idstart = len(atoms) + 1

    pos = atoms.positions
    nop = len(pos)
    ids = atoms.ids
    head = [x for x in range(len(pos))]
    ghosts = [False for x in range(nop)]

    #all the other keys
    datadict = {key: atoms[key][:atoms.nreal] for key in atoms.keys()}
    del datadict['positions']
    del datadict['ids']
    del datadict['head']
    del datadict['ghost']

    #prepopulate
    for d in range(3):
        pos_list = []
        new_id_list = []
        ghost_list = []
        for i in range(1, repetitions[d]):
            npos = copy.copy(pos)
            npos = npos + i*box[d]
            pos_list.append(npos)
            new_ids = [idstart+i for i in range(len(pos))]
            new_id_list.append(new_ids)
            ghost_list.append([ghost for x in range(len(pos))])
            #change id start
            idstart = idstart + len(new_ids)

        pos = np.concatenate((pos, *pos_list))
        ids = np.concatenate((ids, *new_id_list))
        ghosts = np.concatenate((ghosts, *ghost_list))
        #generate new ids
        
        head = np.concatenate((head, np.tile(head, len(pos_list))))

        for key in datadict.keys():
            datadict[key] = np.concatenate((datadict[key], np.tile(datadict[key], len(pos_list))))

    atoms["positions"] = pos
    atoms["ids"] = ids
    atoms["head"] = head
    atoms["ghost"] = ghosts

    for key in datadict.keys():
        atoms[key] = datadict[key]

    if scale_box:
        box[0] = repetitions[0]*np.array(box[0])
        box[1] = repetitions[1]*np.array(box[1])
        box[2] = repetitions[2]*np.array(box[2])
    if ghost:
        system.ghosts_created = True

    if return_atoms:
        return atoms, box

    else:
        system.box = box
        system.atoms = atoms
        return system


def embed_in_cubic_box(system, input_box=None, 
    return_box=False):
    """
    Embedded the triclinic box in a cubic box
    
    Parameters
    ----------
    system : pyscal System
        input system

    input_box: optional
        if specified, this box is used instead of the
        system simulation box

    return_box: bool, optional
        if specified, return the box instead of assigning
        to the system. Default False

    Returns
    -------
    system: 
        system object with the modified box
        only returned if `return_box` is False

    box:
        the cubic simulation box
        only returned if `return_box` is True
    """
    if input_box is None:
        box = copy.copy(system.box)
        backupbox = copy.copy(box)
    else:
        box = copy.copy(input_box)
        backupbox = copy.copy(input_box)

    a = np.array(box[0])
    b = np.array(box[1])
    c = np.array(box[2])

    cosa = np.dot(b, c)/(np.linalg.norm(b)*np.linalg.norm(c))
    cosb = np.dot(c, a)/(np.linalg.norm(c)*np.linalg.norm(a))
    cosc = np.dot(a, b)/(np.linalg.norm(a)*np.linalg.norm(b))

    lx = np.linalg.norm(a)
    xy = np.linalg.norm(b)*cosc
    xz = np.linalg.norm(c)*cosb
    ly = np.sqrt(np.linalg.norm(b)*np.linalg.norm(b) - xy*xy)
    yz = (np.linalg.norm(b)*np.linalg.norm(c)*cosa - xy*xz)/ly
    lz = np.sqrt(np.linalg.norm(c)*np.linalg.norm(c) - xz*xz - yz*yz)

    xlo = ylo = zlo = 0
    xhi = lx
    yhi = ly
    zhi = lz

    xlo_bound = xlo + min(0.0,xy,xz,xy+xz)
    xhi_bound = xhi + max(0.0,xy,xz,xy+xz)
    ylo_bound = ylo + min(0.0,yz)
    yhi_bound = yhi + max(0.0,yz)
    zlo_bound = zlo
    zhi_bound = zhi

    newbox = np.array([[xhi_bound-xlo_bound, 0, 0], [0, yhi_bound-ylo_bound, 0], [0, 0, zhi_bound-zlo_bound]])
    
    if not return_box:
        system.newbox = newbox
        system.box = newbox
        system.box_backup = backupbox
        system.actual_box = None
        return system
    else:
        return newbox


def extract_cubic_representation(system, repetitions = (3,3,3), return_atoms = False):
    """
    Extract a cubic representation of a given box.

    Parameters
    ----------
    system: pyscal System object
        the input system to be repeated

    repetitions: tuple of ints
        number of times the system is to be rotated in each direction

    return_atoms: bool, optional
        if True, return atoms instead of adding them to the system. Default False
    
    Returns
    -------
    system: pyscal System object
        the system with repetitions. Only returned if `return_atoms` is False.
    
    atoms: Atoms object
        only returned if `return_atoms` is True.
    

    Notes
    -----
    This algorithm is not guaranteed to work. Althought it can extract a cubic box,
    there is no guarantee that the extracted box is actually a unit cell.
    """
    atoms, box = pad_repeat(system.atoms, system.box, repetitions, ghost=False)
    
    def _is_in_bound(pos, lo, hi):
        c1 = np.where(pos >= lo, True, False)
        c2 = np.where(pos < hi, True, False)
        return c1*c2

    bounds = []
    seed = 0

    for search_dir, plane in enumerate([[1, 2], [0, 2], [0, 1]]):
        anchor_type = atoms.types[seed]
        anchor = atoms.positions[seed]

        plane_distance_1 = np.abs(atoms.positions[:,plane[0]] - anchor[plane[0]])
        plane_distance_2 = np.abs(atoms.positions[:,plane[1]] - anchor[plane[1]])
        indices_1 = [count for count, p in enumerate(plane_distance_1) if p < 1e-5]
        indices_2 = [count for count, p in enumerate(plane_distance_2) if p < 1e-5]
        indices = list(set(indices_1).intersection(indices_2))
        indices = [index for index in indices if atoms.types[index] == anchor_type]
        distances = distance.cdist([anchor], atoms.positions[indices])[0]
        sorted_args = np.argsort(distances)
        distances = np.array(distances)[sorted_args][1:]
        indices = np.array(indices)[sorted_args][1:]

        if len(indices) == 0:
            raise ValueError(f'Could not find cubic representation, please increase repetitions and try!')
        closest_atom = atoms.positions[indices][0]
        bounds.append([anchor[search_dir], closest_atom[search_dir]])

    #now filter out all atoms which fall in the bounds
    x_atoms = _is_in_bound(atoms.positions[:,0], min(bounds[0]), max(bounds[0]))
    y_atoms = _is_in_bound(atoms.positions[:,1], min(bounds[1]), max(bounds[1]))
    z_atoms = _is_in_bound(atoms.positions[:,2], min(bounds[2]), max(bounds[2]))
    is_needed = x_atoms*y_atoms*z_atoms

    #now get those atoms
    #delete unneeded ones
    unneeded = [count for count in range(len(atoms.positions)) if not is_needed[count]]
    atoms.delete(indices=unneeded)
    box = []
    box.append([max(bounds[0])-min(bounds[0]), 0, 0])
    box.append([0, max(bounds[1])-min(bounds[1]), 0])
    box.append([0, 0, max(bounds[2])-min(bounds[2])])
    
    if return_atoms:
        return atoms, box
    else:
        system.box = box
        system.atoms = atoms
        system.modify.remap_to_box()
        return system


def remap_to_box(system):
    """
    Remap the atom to back inside the box.
    
    Parameters
    ----------
    None

    Returns
    -------
    system
    """
    #rot = np.array(system._box).T
    #rotinv = np.linalg.inv(rot)

    for x in range(system.atoms.ntotal):
        pos = pc.remap_atom_into_box(system.atoms["positions"][x], 
            system.triclinic,
            system.rot, 
            system.rotinv, 
            system.boxdims)
        #print(f'{system.atoms["positions"][x]} changed to {pos} ')
        system.atoms["positions"][x] = pos
    return system       



import numpy as np
import pyscal3.csystem as pc


def get_distance(system, pos1, pos2, vector=False):
    """
    Get the distance between two atoms.

    Parameters
    ----------
    pos1 : list
            first atom position
    pos2 : list
            second atom position
    vector: bool, optional
        If True, return the vector between two atoms

    Returns
    -------
    distance : double
            distance between the first and second atom.

    Notes
    -----
    Periodic boundary conditions are assumed by default.
    """
    diff = pc.get_distance_vector(pos1, pos2, system.triclinic,
        system.rot, system.rotinv, system.boxdims)
    dist = np.linalg.norm(diff)
            
    if vector:
        return dist, diff
    else:
        return dist

def reset_neighbors(system):
    """
    Reset the neighbors of all atoms in the system.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Notes
    -----
    It is used automatically when neighbors are recalculated.

    """
    system.atoms["neighbors"] = []
    system.atoms["neighbordist"] = []
    system.atoms["temp_neighbors"] = []
    system.atoms["temp_neighbordist"] = []
    system.atoms["neighborweight"] = []
    system.atoms["diff"] = []
    system.atoms["r"] = []
    system.atoms["theta"] = []
    system.atoms["phi"] = []
    system.atoms["cutoff"] = []
    system.neighbors_found = False

    
    mapdict = {}
    mapdict["neighbors"] = {}
    mapdict["neighbors"]["index"] = "neighbors"
    mapdict["neighbors"]["distance"] = "neighbordist"
    mapdict["neighbors"]["weight"] = "neighborweight"
    mapdict["neighbors"]["displacement"] = "diff"
    mapdict["neighbors"]["cutoff"] = "cutoff"

    mapdict["neighbors"]["angle"] = {}
    mapdict["neighbors"]["angle"]["polar"] = "theta"
    mapdict["neighbors"]["angle"]["azimuthal"] = "phi"

    mapdict["neighbors"]["temporary"] = {}
    mapdict["neighbors"]["temporary"]["index"] = "temp_neighbors"
    mapdict["neighbors"]["temporary"]["distance"] = "temp_neighbordist"

    system.atoms._add_attribute(mapdict)


	


import pyscal3.traj_process as ptp

def read_inputfile(system, filename, format="lammps-dump", 
                                        compressed = False, customkeys=None):
    """

    Read input file that contains the information of system configuration.

    Parameters
    ----------
    filename : string
        name of the input file.

    format : {'lammps-dump', 'poscar', 'ase', 'mdtraj'}
        format of the input file, in case of `ase` the ASE Atoms object

    compressed : bool, optional
        If True, force to read a `gz` compressed format, default False.

    customkeys : list
        A list containing names of headers of extra data that needs to be read in from the
        input file.

    Returns
    -------
    None

    Notes
    -----
    `format` keyword specifies the format of the input file. Currently only
    a `lammps-dump` and `poscar` files are supported.  Additionaly, the widely
    use Atomic Simulation environment (https://wiki.fysik.dtu.dk/ase/ase/ase.html).
    mdtraj objects (http://mdtraj.org/1.9.3/) are also supported by using the keyword
    `'mdtraj'` for format. Please note that triclinic boxes are not yet supported for
    mdtraj format.
    Atoms object can also be used directly. This function uses the
    :func:`~pyscal.traj_process` module to process a file which is then assigned to system.

    `compressed` keyword is not required if a file ends with `.gz` extension, it is
    automatically treated as a compressed file.

    Triclinic simulation boxes can also be read in.

    If `custom_keys` are provided, this extra information is read in from input files if
    available. This information is can be accessed directly as `self.atoms['customkey']`


    """
    system.neighbors_found = False
    system.neighbor_method = None
    system.ghosts_created = False
    system.actual_box = None

    #box parameters
    system.rot = [[0,0,0], [0,0,0], [0,0,0]]
    system.rotinv = [[0,0,0], [0,0,0], [0,0,0]]
    system.boxdims = [0,0,0]
    system.triclinic = 0

    atoms, box = ptp.read_file(filename, format=format, 
                                compressed=compressed, customkeys=customkeys,)
    system.box = box
    system.atoms = atoms

def to_file(system, outfile, format='lammps-dump', customkeys=None, customvals=None,
            compressed=False, timestep=0, species=None):
    """
    Save the system instance to a trajectory file.

    Parameters
    ----------
    outfile : string
        name of the output file

    format : string, {'lammps-dump', 'lammps-data', 'poscar'}
        format of the output file, default `lammps-dump`
        Currently only `lammps-dump` format is supported.

    customkeys : list of strings, optional
        a list of extra atom wise values to be written in the output file.

    customvals : list or list of lists, optional
        If `customkey` is specified, `customvals` take an array of the same length
        as number of atoms, which contains the values to be written out.

    compressed : bool, optional
        If true, the output is written as a compressed file.

    timestep : int, optional
        timestep to be written to file. default 0

    species : None, optional
        species of the atoms. Required if any format other than 'lammps-dump' is used. Required
        for convertion to ase object.

    Returns
    -------
    None

    Notes
    -----
    `to_file` method can handle a number of file formats. The most customizable format is the
    `lammps-dump` which can take a custom options using customkeys and customvals. customkeys
    will be the header written to the dump file. It can be any Atom attribute, any property
    stored in custom variable of the Atom, or calculated q values which can be given by `q4`,
    `aq4` etc. External values can also be provided using `customvals` option. `customvals` array
    should be of the same length as the number of atoms in the system.

    For all other formats, ASE is used to write out the file, and hence the `species` keyword
    needs to be specified. If initially, an ASE object was used to create the System, `species`
    keyword will already be saved, and need not be specified. In other cases, `species` should
    be a list of atomic species in the System. For example `["Cu"]` or `["Cu", "Al"]`, depending
    on the number of species in the System. In the above case, atoms of type 1 will be mapped to
    Cu and of type 2 will be mapped to Al. For a complete list of formats that ASE can handle,
    see `here <https://wiki.fysik.dtu.dk/ase/ase/io/io.html>`_ . 
    """

    ptp.write_file(system, outfile, format = format,
        compressed = compressed, customkeys = customkeys, customvals = customvals,
        timestep = timestep, species = species)


import pyscal3.core as pc
import os
import numpy as np
from ase.build import bulk

def test_voronoi_props():
    nx = 5
    sys = pc.System.create.lattice.bcc(repetitions = [nx, nx, nx], lattice_constant=3.127)
    sys.find.neighbors(method="voronoi")

    assert sys.atoms.voronoi.vertex.numbers[0][0] == 6
    assert (sys.atoms.voronoi.vertex.positions[0][0][0]+1.5635 < 1E-4)
    assert (sys.atoms.voronoi.volume[0]-15.288104691499992 < 1E-5)
    assert (sys.atoms.voronoi.face.perimeters[0][0]-6.6333687143110005 < 1E-5)
    assert sys.atoms.voronoi.face.vertices[0][0] == 6

def test_voronoi_vector():
    sys = pc.System.create.lattice.fcc(repetitions=(4,4,4))
    sys.find.neighbors(method='voronoi')
    sys.calculate.voronoi_vector()
    assert sys.atoms.voronoi.vector[0][1] == 12

#def test_voronoi_vertices():
#    nx = np.random.randint(1, 10)
#    nverts = (nx**3*12)
#    sys = Structure().lattice.bcc(repetitions = [nx, nx, nx])
#    sys.find_neighbors(method='voronoi', cutoff=0.1)
#    assert len(sys.atoms.voronoi.vertex.unique_positions) == nverts


import pytest
import os
import numpy as np
import pyscal3.core as pc

def test_complex_system():
    sys = pc.System('tests/files/cluster.dump')
    sys.find.neighbors(method='cutoff', cutoff=3.63)
    assert 176 == sys.find.solids(bonds=6, threshold=0.5, avgthreshold=0.6, cluster=True)

def test_cluster():
    sys = pc.System('tests/files/cluster.dump')
    sys.find.neighbors(method='cutoff', cutoff=3.63)
    sys.find.solids(cluster=False)
    val = sys.find.clusters(sys.atoms.steinhardt.order.sij.solid, largest = True)
    assert 176 == val

def test_cluster_cutoff():
    sys = pc.System('tests/files/cluster.dump')
    sys.find.neighbors(method='cutoff', cutoff=3.63)
    sys.find.solids(cluster=False)
    val = sys.find.clusters(sys.atoms.steinhardt.order.sij.solid, largest = True, cutoff=3.63)
    assert 176 == val

def test_system_nucsize_fraction():
    #create some atoms
    sys = pc.System.create.lattice.bcc(repetitions = [2,2,2], lattice_constant=3.20)    

    #test that atoms are set properly
    assert sys.natoms == 16

    #now calculate nucsize
    sys.find.neighbors(method='cutoff', cutoff=3.63)
    assert 16 == sys.find.solids(bonds=0.8, threshold=0.5, avgthreshold=0.6, cluster=True)


import pyscal3.core as pc
import os
import numpy as np
from ase.build import bulk

def test_system_init():
	sys = pc.System("tests/files/conf.dump", customkeys=["vx", "vy"])
	assert sys.natoms == 500
	assert sys.atoms.vx[0] == '0.0394436'

	sys.read.file("tests/files/conf.dump.gz")
	assert sys.natoms == 500

	sys.read.file("tests/files/conf.bcc.scaled.dump")
	assert sys.natoms == 2

	sys.read.file("tests/files/POSCAR", format="poscar")
	assert sys.natoms == 42

	sys = pc.System("tests/files/conf.dump", customkeys=["vx", "vy"])
	sys.write.file("test.dump", customkeys=["vx"])

	sys = pc.System("test.dump", customkeys=["vx"])
	assert sys.natoms == 500
	assert sys.atoms.vx[0] == '0.0394436'

	sys = pc.System("tests/files/conf.dump", customkeys=["vx", "vy"])
	aseobj = sys.write.ase(species=["Au"])
	assert np.sum(sys.atoms.positions[0]-aseobj.positions[0]) < 1E-5


import pyscal3.core as pc
import os
import numpy as np

def test_masking():
	pass
	#atoms, box = pcs.make_crystal(structure='b2', repetitions=(6,6,6), lattice_constant=3.127)
	#sys = pc.System()
	#sys.box = box
	#sys.atoms = atoms

	#def _condition1(atom):
	#	if atom.species[0] == 1:
	#		return True
	#	return False

	#def _condition2(atom):
	#	if atom.species[0] == 1:
	#		return False
	#	return True

	#sys.apply_mask(condition=_condition1, mask_type="secondary")
	#sys.apply_mask(condition=_condition2, mask_type="primary")	
	#sys.find_neighbors(method="cutoff", cutoff=3.6)
	#assert len(sys.atoms.neighbors.index[0]) == 0
	#sys.remove_mask(mask_type="all")

	#sys.apply_mask(condition=_condition2, mask_type="primary")
	#sys.find_neighbors(method="cutoff", cutoff=3.6)
	#assert len(sys.atoms.neighbors.index[1]) == 7
	#sys.remove_mask(mask_type="all")

	#sys.find_neighbors(method="cutoff", cutoff=3.6)
	#assert len(sys.atoms.neighbors.index[0]) == 14
	#assert len(sys.atoms.neighbors.index[1]) == 14



import pytest
import numpy as np
import pyscal3.core as pc

def test_chiparamsbcc():
    sys = pc.System.create.lattice.bcc(repetitions = [3, 3, 3], lattice_constant=4)
    sys.find.neighbors(method='cutoff', cutoff=0)
    sys.calculate.chi_params()
    chip2 = [3, 0, 0, 0, 36, 12, 0, 36, 0]
    assert np.sum(np.array(sys.atoms.angular_parameters.chi_params[2])-np.array(chip2)) == 0


def test_chiparamsfcc():
    sys = pc.System.create.lattice.fcc(repetitions = [5, 5, 5], lattice_constant=4)
    sys.find.neighbors(method='cutoff', cutoff=0)
    sys.calculate.chi_params()
    chip2 =  [6, 0, 0, 0, 24, 12, 0, 24, 0]
    assert np.sum(np.array(sys.atoms.angular_parameters.chi_params[2])-np.array(chip2)) == 0

def test_chiparamsdia():
    sys = pc.System.create.lattice.diamond(repetitions = [3, 3, 3], lattice_constant=4)
    sys.find.neighbors(method='cutoff', cutoff=0)
    sys.calculate.chi_params()
    chip2 =  [0, 0, 0, 0, 6, 0, 0, 0, 0]
    assert np.sum(np.array(sys.atoms.angular_parameters.chi_params[2])-np.array(chip2)) == 0



import pytest
import os,sys,inspect
import numpy as np
import pyscal3.core as pc

def test_cs_ges():
    sys = pc.System.create.lattice.bcc(repetitions = [7, 7, 7], lattice_constant=4.00)
    q = sys.calculate.centrosymmetry(nmax=8)
    assert np.round(np.mean(np.array(q)), decimals=2) == 0.00

    q = sys.calculate.centrosymmetry(nmax=4)
    assert np.round(np.mean(np.array(q)), decimals=2) > 0.00



import numpy as np

def calculate_vorovector(system, edge_cutoff=0.05, area_cutoff=0.01, edge_length=False):
    """
    get the voronoi structure identification vector.

    Parameters
    ----------

    edge_cutoff : float, optional
        cutoff for edge length. Default 0.05.


    area_cutoff : float, optional
        cutoff for face area. Default 0.01.

    edge_length : bool, optional
        if True, a list of unrefined edge lengths are returned. Default false.

    Returns
    -------
    vorovector : array like, int
        array of the form (n3, n4, n5, n6)

    Notes
    -----
    Returns a vector of the form `(n3, n4, n5, n6)`, where `n3` is the number
    of faces with 3 vertices, `n4` is the number of faces with 4
    vertices and so on. This can be used to identify structures [1] [2].

    The keywords `edge_cutoff` and `area_cutoff` can be used to tune the values to minimise
    the effect of thermal distortions. Edges are only considered in the analysis if the
    `edge_length/sum(edge_lengths)` is at least `edge_cutoff`. Similarly, faces are only
    considered in the analysis if the  `face_area/sum(face_areas)` is at least `face_cutoff`.

    References
    ----------
    .. [1] Finney, JL, Proc. Royal Soc. Lond. A 319, 1970
    .. [2] Tanemura, M, Hiwatari, Y, Matsuda, H,Ogawa, T, Ogita, N, Ueda, A. Prog. Theor. Phys. 58, 1977

    """
    complete_edge_lengths = []
    vorovectors = []

    for x in range(len(system.atoms['positions'])):
        st = 1
        refined_edges = []
        edge_lengths = []
        for vno in system.atoms['face_vertices'][x]:
            vphase = system.atoms['vertex_numbers'][x][st:st+vno]
            edgecount = 0
            dummy_edge_lengths = []
            #now calculate the length f each edge
            for i in range(-1, len(vphase)-1):
                #get pairs of indices
                #verts are i, i+1
                ipos = system.atoms['vertex_vectors'][x][vphase[i]*3:vphase[i]*3+3]
                jpos = system.atoms['vertex_vectors'][x][vphase[i+1]*3:vphase[i+1]*3+3]

                #now calculate edge length
                edgeln = np.sqrt((ipos[0]-jpos[0])**2 + (ipos[1]-jpos[1])**2 + (ipos[2]-jpos[2])**2)
                dummy_edge_lengths.append(edgeln)

            edge_lengths.append(dummy_edge_lengths)
            st += (vno+1)

        #now all the edge lengths are saved
        for c, ed in enumerate(edge_lengths):
            #normalise the edge lengths
            norm = (ed/np.sum(ed))
            #apply face area cutoff
            if (system.atoms['neighborweight'][x][c] > area_cutoff):
                #check for edge length cutoff
                edgecount = len([cc for cc,x in enumerate(norm) if x > edge_cutoff])
                refined_edges.append(edgecount)

        #now loop over refined edges and collect n3, n4, n5, n6
        vorovector = [0, 0, 0, 0]

        for ed in refined_edges:
            if ed == 3:
                vorovector[0] += 1
            elif ed == 4:
                vorovector[1] += 1
            elif ed == 5:
                vorovector[2] += 1
            elif ed == 6:
                vorovector[3] += 1

        complete_edge_lengths.append(edge_lengths)
        vorovectors.append(vorovector)
    
    system.atoms["edge_lengths"] = np.array(complete_edge_lengths)
    system.atoms["vorovector"] = np.array(vorovectors)

    mapdict = {}
    mapdict["voronoi"] ={}
    mapdict["voronoi"]["vector"] = "vorovector"
    mapdict["voronoi"]["face"] = {}
    mapdict["voronoi"]["face"]["edge_lengths"] = "edge_lengths"
    system.atoms._add_attribute(mapdict)




import pytest
import os
import numpy as np
import pyscal3.core as pc


def test_q_3():
    sys = pc.System.create.lattice.bcc(repetitions = [4, 4, 4])

    #sys.get_neighbors(method = 'voronoi')
    sys.find.neighbors(method = 'cutoff', cutoff=0.9)

    q = sys.calculate.steinhardt_parameter(3, averaged=True)
    assert np.round(np.mean(np.array(q)), decimals=2) == 0.00



import numpy as np
import itertools
import warnings
import matplotlib.pyplot as plt
import matplotlib

def create_box_plot(box, origin=[0,0,0]):
    """
    Create traces which correspond to the simulation cell

    Parameters
    ----------
    box : list
        dimensions of the simulation box

    origin : list, optional
        Origin of the simulation box. Default [0, 0, 0]

    Returns
    -------
    traces : list of Scatter3d objects

    """

    try:
        from plotly import graph_objs as go
        import ipywidgets as widgets
    except ImportError:
        print("Install ipyvolume for visualisation")

    box = np.array(box)
    origin = np.array(origin)
    combos = list(itertools.combinations(range(3), 2))
    faces = []
    for combo in combos:
        f1 = [origin, box[combo[0]], box[combo[0]]+box[combo[1]], box[combo[1]], origin]
        s = combo[0] + combo[1]
        t = 3-s
        f2 = [origin + box[t], box[combo[0]]+ box[t],  box[combo[0]]+box[combo[1]]+ box[t], box[combo[1]]+ box[t], origin + box[t]]
        faces.append(np.array(f1))
        faces.append(np.array(f2))
    traces = []
    for face in faces:
        trace = go.Scatter3d(
            x=face[:,0],
            y=face[:,1],
            z=face[:,2],
            mode='lines',
            name='lines',
            line=dict(width=2.0, color='#263238'),
            showlegend=False
        )
        traces.append(trace)
    return traces

def plot_by_selection(sys, radius=10, 
    opacity=1.0):

    try:
        from plotly import graph_objs as go
    except ImportError:
        print("Install plotly for visualisation")

    cFalse = '#b0bec5'
    cTrue = '#ef5350'

    #plotting box
    box = sys.box.copy()
    origin = np.array([0,0,0])
    traces = create_box_plot(box, origin)

    data=go.Scatter3d(
        x=sys.atoms.positions[:,0][sys.atoms.selection],
        y=sys.atoms.positions[:,1][sys.atoms.selection],
        z=sys.atoms.positions[:,2][sys.atoms.selection],
        mode='markers',
        opacity=1.0,
        marker=dict(
            sizemode='diameter',
            sizeref=750,
            size=radius,
            color = cTrue,
            opacity = opacity,
            #colorscale = colorscale,
            #colorbar=dict(thickness=20, title=cmap_title),
            line=dict(width=0.5, color='#455A64'),            
        )
    )

    traces.append(data)

    data=go.Scatter3d(
        x=sys.atoms.positions[:,0][np.logical_not(sys.atoms.selection)],
        y=sys.atoms.positions[:,1][np.logical_not(sys.atoms.selection)],
        z=sys.atoms.positions[:,2][np.logical_not(sys.atoms.selection)],
        mode='markers',
        opacity=1.0,
        marker=dict(
            sizemode='diameter',
            sizeref=750,
            size=radius,
            color = cFalse,
            opacity = opacity,
            #colorscale = colorscale,
            #colorbar=dict(thickness=20),
            line=dict(width=0.5, color='#455A64')
        )
    )

    traces.append(data)

    fig = go.Figure(data=traces)
    fig.update_layout(scene = dict(
                        xaxis_title="",
                        yaxis_title="",
                        zaxis_title="",
                        xaxis = dict(
                             showticklabels=False,
                             showbackground=False,
                             zerolinecolor="#455A64",),
                        yaxis = dict(
                            showticklabels=False,
                            showbackground=False,
                            zerolinecolor="#455A64"),
                        zaxis = dict(
                            showticklabels=False,
                            showbackground=False,
                            zerolinecolor="#455A64",),),
                        width=700,
                        margin=dict(
                        r=10, l=10,
                        b=10, t=10)
                      )
    fig.update_layout(showlegend=False)
    fig['layout'].update(scene=dict(aspectmode="data"))
    return fig.show()


def plot_by_property(sys, colorby, 
    cmap = 'viridis', 
    radius=10, 
    opacity=1.0,
    hide_zero=False):


    try:
        from plotly import graph_objs as go
    except ImportError:
        print("Install plotly for visualisation")
    
    #colorby = colorby.copy().astype(int)

    #sys.apply_selection(ids=ids, indices=indices, condition=condition)
    #colorby = [x for count, x in enumerate(colorby) if sys.atoms.selection[count]]
    
    #colorby = colorby.copy()
    #a = min(colorby)
    #b = max(colorby)
    #prop = (colorby - a)/(b - a)

    box = sys.box.copy()
    origin = np.array([0,0,0])
    traces = create_box_plot(box, origin)

    if hide_zero:
        data=go.Scatter3d(
            x=sys.atoms.positions[:,0][np.where(colorby > 0)],
            y=sys.atoms.positions[:,1][np.where(colorby > 0)],
            z=sys.atoms.positions[:,2][np.where(colorby > 0)],
            mode='markers',
            opacity=1.0,
            marker=dict(
                sizemode='diameter',
                sizeref=750,
                size=radius,
                color = colorby,
                opacity = opacity,
                colorscale = cmap,
                colorbar=dict(thickness=20),
                line=dict(width=0.5, color='#455A64'),            
            )
        )
    else:
        data=go.Scatter3d(
            x=sys.atoms.positions[:,0],
            y=sys.atoms.positions[:,1],
            z=sys.atoms.positions[:,2],
            mode='markers',
            opacity=1.0,
            marker=dict(
                sizemode='diameter',
                sizeref=750,
                size=radius,
                color = colorby,
                opacity = opacity,
                colorscale = cmap,
                colorbar=dict(thickness=20),
                line=dict(width=0.5, color='#455A64'),            
            )
        )

    traces.append(data)

    fig = go.Figure(data=traces)
    fig.update_layout(scene = dict(
                        xaxis_title="",
                        yaxis_title="",
                        zaxis_title="",
                        xaxis = dict(
                             showticklabels=False,
                             showbackground=False,
                             zerolinecolor="#455A64",),
                        yaxis = dict(
                            showticklabels=False,
                            showbackground=False,
                            zerolinecolor="#455A64"),
                        zaxis = dict(
                            showticklabels=False,
                            showbackground=False,
                            zerolinecolor="#455A64",),),
                        width=700,
                        margin=dict(
                        r=10, l=10,
                        b=10, t=10)
                      )
    fig.update_layout(showlegend=False)
    fig['layout'].update(scene=dict(aspectmode="data"))
    #add plot
    #sys.remove_selection()
    return fig.show()


def plot_by_boolean(sys, colorby, 
    color = '#ff7f00', 
    radius=10, 
    opacity=1.0,
    hide_zero=False):


    try:
        from plotly import graph_objs as go
    except ImportError:
        print("Install plotly for visualisation")
    
    #colorby = colorby.copy().astype(int)

    #sys.apply_selection(ids=ids, indices=indices, condition=condition)
    #colorby = [x for count, x in enumerate(colorby) if sys.atoms.selection[count]]
    
    #colorby = colorby.copy()
    #a = min(colorby)
    #b = max(colorby)
    #prop = (colorby - a)/(b - a)

    box = sys.box.copy()
    origin = np.array([0,0,0])
    traces = create_box_plot(box, origin)

    data=go.Scatter3d(
        x=sys.atoms.positions[:,0][np.where(colorby == True)],
        y=sys.atoms.positions[:,1][np.where(colorby == True)],
        z=sys.atoms.positions[:,2][np.where(colorby == True)],
        mode='markers',
        opacity=1.0,
        marker=dict(
            sizemode='diameter',
            sizeref=750,
            size=radius,
            color = color,
            opacity = opacity,
            #colorscale = cmap,
            #colorbar=dict(thickness=20),
            line=dict(width=0.5, color='#455A64'),            
        )
    )

    traces.append(data)

    fig = go.Figure(data=traces)
    fig.update_layout(scene = dict(
                        xaxis_title="",
                        yaxis_title="",
                        zaxis_title="",
                        xaxis = dict(
                             showticklabels=False,
                             showbackground=False,
                             zerolinecolor="#455A64",),
                        yaxis = dict(
                            showticklabels=False,
                            showbackground=False,
                            zerolinecolor="#455A64"),
                        zaxis = dict(
                            showticklabels=False,
                            showbackground=False,
                            zerolinecolor="#455A64",),),
                        width=700,
                        margin=dict(
                        r=10, l=10,
                        b=10, t=10)
                      )
    fig.update_layout(showlegend=False)
    fig['layout'].update(scene=dict(aspectmode="data"))
    #add plot
    #sys.remove_selection()
    return fig.show()



def plot_simple(sys, colors=None,  
    radius=10, 
    opacity=1.0 ):

    try:
        from plotly import graph_objs as go
    except ImportError:
        print("Install plotly for visualisation")
    
    if colors is None:
        colors =[ '#33a02c', '#fb9a99', '#e31a1c',
                 '#a6cee3', '#1f78b4', '#b2df8a', 
                 '#fdbf6f', '#ff7f00', '#cab2d6', 
                 '#6a3d9a', '#ffff99', '#b15928']
    
    comp_ints = sys.atoms.composition_ints
    if len(comp_ints) > len(colors):
        warnings.warn('less colors than number of species, expect repetitions, or provide more colors')
        diff = np.ceil((len(comp_ints)-len(colors))/len(colors))
        colors = colors*diff
    

    box = sys.box.copy()
    origin = np.array([0,0,0])
    traces = create_box_plot(box, origin)

    for key in comp_ints.keys():
        data=go.Scatter3d(
            x=sys.atoms.positions[:,0][sys.atoms.types == key],
            y=sys.atoms.positions[:,1][sys.atoms.types == key],
            z=sys.atoms.positions[:,2][sys.atoms.types == key],
            mode='markers',
            opacity=1.0,
            marker=dict(
                sizemode='diameter',
                sizeref=750,
                size=radius,
                color = colors[key-1],
                opacity = opacity,
                #colorbar=dict(thickness=20, title=cmap_title),
                line=dict(width=0.5, color='#455A64'),            
            )
        )

        traces.append(data)

    fig = go.Figure(data=traces)
    fig.update_layout(scene = dict(
                        xaxis_title="",
                        yaxis_title="",
                        zaxis_title="",
                        xaxis = dict(
                             showticklabels=False,
                             showbackground=False,
                             zerolinecolor="#455A64",),
                        yaxis = dict(
                            showticklabels=False,
                            showbackground=False,
                            zerolinecolor="#455A64"),
                        zaxis = dict(
                            showticklabels=False,
                            showbackground=False,
                            zerolinecolor="#455A64",),),
                        width=700,
                        margin=dict(
                        r=10, l=10,
                        b=10, t=10)
                      )
    fig.update_layout(showlegend=False)
    fig['layout'].update(scene=dict(aspectmode="data"))
    return fig.show()

    



import yaml
import os

doc_fields = ["doc", "url"]

def read_yaml(filename):
    with open(filename, 'r') as fin:
        data = yaml.safe_load(fin)
        return data

def _get_doc_from_key(keydict):
    url =  keydict["url"] if "url" in keydict.keys() else None

    doc = f"""
    {keydict["doc"]}
    url: {url}
    """
    return doc

class MyList(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class AttrSetter:
    """
    Class which enables tab completed, contextual attributes
    """    
    def __init__(self):
        self._map_dict = {}
        self.head = None

    def __dir__(self):
        """
        Gives tab completion
        """
        return list(self._map_dict.keys())

    def __repr__(self):
        """
        String representation
        """
        return ", ".join(list(self._map_dict.keys()))

    def __getattr__(self, key):
        """
        Calls attributes from the class when general lookup fails. There are four main types of
        calls that can occur.

        Normal attribute: this function does not get called
        key which is present in `_map_dict`: filter and call it
        key which add ons: for example `position_selected` or `position_not_selected`
        key does not exist: raise `AttributeError`
        """
        if key in self._map_dict.keys():
            if self.head is None:
                self.head = self
            return self._check_filter(key)

        #now if it doesnt exists
        stripped_key, search_key, which_type, reverse = self._has_addons(key)
        #print(stripped_key, search_key, which_type, reverse)
        if (stripped_key == key):
            #this didnt work
            raise AttributeError(f"Attribute {stripped_key} not found")

        else:
            return self._get_key_on_condition(stripped_key, search_key, which_type, reverse)


    def _check_filter(self, key):
        """
        This checks for non-filtered keywords
        """
        if isinstance(self._map_dict[key], str):
            val_raw = self._map_dict[key].split("/")
            if val_raw[-1] == "nofilter":
                return self.head[val_raw[0]]
            else:
                return self.head[self._map_dict[key]][:self.head.nreal]
        else:
            return self._map_dict[key]

    def _has_addons(self, key):
        raw = key.split("_")
        which_type = None
        reverse = False
        search_key = None

        if "for" in raw:
            index = raw.index("for")
            which_type = raw[index+1] if raw[index+1] in ["real", "ghost", "all"] else None
            #now remove these keys
            del raw[index+1]
            del raw[index]

        if "with" in raw:
            index = raw.index("with")
            with_raw = raw[index+1:]

            if with_raw[0] in ["un", "non", "de", "not", "no"]:
                reverse = True
            if with_raw[-1] in ["selection", "select", "selected", "condition"]:
                search_key = "condition"
            elif with_raw[-1] in ["ghost"]:
                search_key = "ghost"
            elif with_raw[-1] in ["real"]:
                search_key = "real"
            elif with_raw[-1] in ["all"]:
                search_key = "all"
            elif with_raw[-1] in ["maskprimary"]:
                search_key = "mask_1"
            elif with_raw[-1] in ["masksecondary"]:
                search_key = "mask_2"
            
            if len(with_raw) == 2:
                del raw[index+2]
            del raw[index+1]
            del raw[index]

        stripped_key = "_".join(raw)
        return stripped_key, search_key, which_type, reverse

    def _get_key_on_condition(self, stripped_key, search_key, which_type, reverse):
        #print(stripped_key, search_key, which_type, reverse)
        search_val = None
        if stripped_key in self._map_dict.keys():
            if which_type is None:
                which_type = "all"
            if which_type == "all":
                stripped_val = self.head[self._map_dict[stripped_key]]
                if search_key is not None:
                    search_val = self.head[self._map_dict[search_key]]
            elif which_type == "real":
                stripped_val = self.head[self._map_dict[stripped_key]][:self.head.nreal]
                if search_key is not None:
                    search_val = self.head[self._map_dict[search_key]][:self.head.nreal]
            elif which_type == "ghost":
                stripped_val = self.head[self._map_dict[stripped_key]][self.head.nreal:]
                if search_key is not None:
                    search_val = self.head[self._map_dict[search_key]][self.head.nreal:]

            #we have gathered stripped_val and search_val
            if search_val is not None:
                if reverse:
                    stripped_val = [x for count, x in enumerate(stripped_val) if not search_val[count]]
                else:
                    stripped_val = [x for count, x in enumerate(stripped_val) if search_val[count]]

            return stripped_val
        
        else:
            raise AttributeError("Attribute not found")

    def _add_attribute(self, indict, head=None):
        if head is None:
            head = self
        self.head = head

        for key, val in indict.items():
            if isinstance(val, dict):
                if key not in self._map_dict.keys():
                    self._map_dict[key] = AttrSetter()

                self._map_dict[key]._add_attribute(val, head=head)
            else:
                self._map_dict[key] = val

def generate_input_template():
    indict = {"attribute_name":{
        "value": None,
        "doc": None,
        "unit": None,
        "url": None,
    }}
    return indict

def _is_sub_input(indict):
    if "value" in indict.keys():
        return False
    return True

def _create_getter(key):
    def _getter(self):
        return getattr(self, f"_{key}")
    return _getter

def _create_setter(key):
    def _setter(self, value):
        setattr(self, f"_{key}", value)
    return _setter

def _get_doc_from_key(keydict):
    url =  keydict["url"] if "url" in keydict.keys() else None
    unit =  keydict["unit"] if "unit" in keydict.keys() else None
    value =  keydict["value"] if "value" in keydict.keys() else None
    doc =  keydict["doc"] if "doc" in keydict.keys() else None
    typeinfo =  keydict["type"] if "type" in keydict.keys() else None
    optional = False if value is None else True    

    doc = f"""
    {doc}  
    Unit: {unit}
    Default: {value}
    url: {url}
    Optional: {optional}
    type: {typeinfo}
    """
    return doc

class PropertySetter(type):
    def __new__(cls, name, bases, dct, input_params):
        for key, val in input_params.items():
            if not _is_sub_input(val):
                dct[key] = property(fget=_create_getter(key), fset=_create_setter(key), doc=_get_doc_from_key(val))
        x = super().__new__(cls, name, bases, dct)
        return x
    
    def __init__(self, name, bases, dct, input_params):
        for key, val in input_params.items():
            if not _is_sub_input(val): 
                setattr(self, f"_{key}", val["value"])
        
class DocumentedKeywords:
    def _create_input_tree(self, indict, name=None, key=None):
        if name is None:
            name = "input"
            
        #create the main input
        _cmain = PropertySetter("CMain", (), {}, input_params=indict)
        
        if key is None:
            setattr(self, name, _cmain())
        else:
            head = getattr(self, name)
            setattr(head, key, _cmain())
        
        for key, val in indict.items():
            if _is_sub_input(val):
                self._create_input_tree(val, name=name, key=key)
                name = key

                




import numpy as np
import pyscal3.csystem as pc

def find_neighbors(system, method='cutoff', cutoff=0, threshold=2, 
        voroexp=1, padding=1.2, nlimit=6, 
        cells=None, nmax=12, assign_neighbor=True):
    """

    Find neighbors of all atoms in the :class:`~pyscal.core.System`.

    Parameters
    ----------
    method : {'cutoff', 'voronoi', 'number'}
        `cutoff` method finds neighbors of an atom within a specified or adaptive cutoff distance from the atom.
        `voronoi` method finds atoms that share a Voronoi polyhedra face with the atom. Default, `cutoff`
        `number` method finds a specified number of closest neighbors to the given atom. Number only populates
        

    cutoff : { float, 'sann', 'adaptive'}
        the cutoff distance to be used for the `cutoff` based neighbor calculation method described above.
        If the value is specified as 0 or `adaptive`, adaptive method is used.
        If the value is specified as `sann`, sann algorithm is used.

    threshold : float, optional
        only used if ``cutoff=adaptive``. A threshold which is used as safe limit for calculation of cutoff.

    voroexp : int, optional
        only used if ``method=voronoi``. Power of the neighbor weight used to weight the contribution of each atom towards
        Steinhardt parameter values. Default 1.

    padding : double, optional
        only used if ``cutoff=adaptive`` or ``cutoff=number``. A safe padding value used after an adaptive cutoff is found. Default 1.2.

    nlimit : int, optional
        only used if ``cutoff=adaptive``. The number of particles to be considered for the calculation of adaptive cutoff.
        Default 6.
    
    cells : bool, optional
        If True, always use cell lists. Default None.

    nmax : int, optional
        only used if ``cutoff=number``. The number of closest neighbors to be found for each atom. Default 12
    

    Returns
    -------
    None

    Raises
    ------
    RuntimeWarning
        raised when `threshold` value is too low. A low threshold value will lead to 'sann' algorithm not converging
        when finding a neighbor. This function will try to automatically increase `threshold` and check again.

    RuntimeError
        raised when neighbor search was unsuccessful. This is due to a low `threshold` value.

    Notes
    -----
    This function calculates the neighbors of each particle. There are several ways to do this. A complete description of
    the methods can be `found here <https://pyscal.readthedocs.io/en/latest/nearestneighbormethods.html>`_.

    Method cutoff and specifying a cutoff radius uses the traditional approach being the one in which the neighbors of an atom
    are the ones that lie in the cutoff distance around it.

    In order to reduce time during the distance sorting during thefind_neighbors adaptive methods, pyscal sets an initial guess for a cutoff distance.
    This is calculated as,

    .. math:: r_{initial} = threshold * (simulation~box~volume/ number~of~particles)^{(1/3)}

    threshold is a safe multiplier used for the guess value and can be set using the `threshold` keyword.

    In Method cutoff, if ``cutoff='adaptive'``, an adaptive cutoff is found during runtime for each atom [1].
    Setting the cutoff radius to 0 also uses this algorithm. The cutoff for an atom i is found using,

    .. math:: r_c(i) = padding * ((1/nlimit) * \sum_{j=1}^{nlimit}(r_{ij}))

    padding is a safe multiplier to the cutoff distance that can be set through the keyword `padding`. `nlimit` keyword sets the
    limit for the top nlimit atoms to be taken into account to calculate the cutoff radius.

    In Method cutoff, if ``cutoff='sann'``, sann algorithm is used [2]. There are no parameters to tune sann algorithm.

    The second approach is using Voronoi polyhedra which also assigns a weight to each neighbor in the ratio of the face area between the two atoms.
    Higher powers of this weight can also be used [3]. The keyword `voroexp`
    can be used to set this weight.
    
    If method is `number`, instead of using a cutoff value for finding neighbors, a specified number of closest atoms are
    found. This number can be set through the argument `nmax`.

    If `cells` is None, cell lists are used if number of atoms are higher than 2500. If True, cell lists are always used.

    .. warning::

        Adaptive and number cutoff uses a padding over the intial guessed "neighbor distance". By default it is 2. In case
        of a warning that ``threshold`` is inadequate, this parameter should be further increased. High/low value
        of this parameter will correspond to the time taken for finding neighbors.

    References
    ----------
    .. [1] Stukowski, A, Model Simul Mater SC 20, 2012
    .. [2] van Meel, JA, Filion, L, Valeriani, C, Frenkel, D, J Chem Phys 234107, 2012
    .. [3] Haeberle, J, Sperl, M, Born, P, arxiv 2019

    """
    #first reset all neighbors
    system.reset_neighbors()
    system.filter = 0

    if threshold < 1:
        raise ValueError("value of threshold should be at least 1.00")

    if cells is None:
        cells = (system.natoms > 250)

    if method == 'cutoff':            
        if cutoff=='sann':    
            finished = False
            for i in range(1, 10):
                finished = pc.get_all_neighbors_sann(system.atoms, 0.0, 
                    system.triclinic, system.rot, system.rotinv,
                    system.boxdims, threshold*i, cells)
                if finished:
                    if i>1:
                        warnings.warn("found neighbors with higher threshold than default/user input")
                    break
                warnings.warn("Could not find sann cutoff. trying with a higher threshold", RuntimeWarning)
            else:
                raise RuntimeError("sann cutoff could not be converged. This is most likely, \
                    due to a low threshold value. Try increasing it.")
        
        elif cutoff=='adaptive' or cutoff==0:
            finished = pc.get_all_neighbors_adaptive(system.atoms, 0.0,
                system.triclinic, system.rot, system.rotinv,
                system.boxdims, threshold, nlimit, padding, cells)
            if not bool(finished):
                raise RuntimeError("Could not find adaptive cutoff")
        else:
            if cells:
                pc.get_all_neighbors_cells(system.atoms, cutoff,
                    system.triclinic, system.rot, system.rotinv, system.boxdims)
            else:
                pc.get_all_neighbors_normal(system.atoms, cutoff,
                    system.triclinic, system.rot, system.rotinv, system.boxdims)

    elif method == 'number':
        finished = pc.get_all_neighbors_bynumber(system.atoms, 0.0, 
            system.triclinic, system.rot, system.rotinv,
            system.boxdims, threshold, nmax, cells, assign_neighbor)
        if not finished:
            raise RuntimeError("Could not find enough neighbors - try increasing threshold")

    
    elif method == 'voronoi':
        clean_vertices = (cutoff>0)
        #CLEANING is TURNED OFF
        #if not clean_vertices:
            #copy the simulation cell
        backupbox = system._box.copy()
        if system.triclinic:
            if not system.ghosts_created:
                atoms, box = operations.repeat(system, (2, 2, 2), atoms=system.atoms, ghost=True, return_atoms=True)
                system.actual_box = system.box.copy()
                system.internal_box = box
                system._atoms = atoms
                system = system.embed_in_cubic_box()
        pc.get_all_neighbors_voronoi(system.atoms, 0.0,
            system.triclinic, system.rot, system.rotinv,
            system.boxdims, voroexp)

        if system.triclinic:
            system._box = backupbox

        #now clean up
        #else:
        #    real_atomdict = {"positions":copy.copy(system.atoms.positions_for_all), 
        #     "ghost":copy.copy(system.atoms.positions_for_ghost)}
            #we need to call the method
            #this means alles good
        #    if system.actual_box is None:
        #        if system.triclinic:
        #            new_box = system.embed_in_cubic_box(inputbox=system._box, return_box=True)
        #            rot = np.array(new_box).T
        #            rotinv = np.linalg.inv(rot)
        #        else:
        #            new_box = system._box
        #            rot = [[0,0,0], [0,0,0], [0,0,0]]
        #            rotinv = [[0,0,0], [0,0,0], [0,0,0]]
        #    #ghosts are present
        #    else:
        #        if system.triclinic:
        #            new_box = system.embed_in_cubic_box(inputbox=system.actual_box, return_box=True)
        #            rot = np.array(new_box).T
        #            rotinv = np.linalg.inv(rot)
        #        else:
        #            new_box = system.actual_box
        #            rot = [[0,0,0], [0,0,0], [0,0,0]]
        #            rotinv = [[0,0,0], [0,0,0], [0,0,0]]

        #    boxdims = [0,0,0]
        #    boxdims[0] = np.sum(np.array(new_box[0])**2)**0.5
        #    boxdims[1] = np.sum(np.array(new_box[1])**2)**0.5
        #    boxdims[2] = np.sum(np.array(new_box[2])**2)**0.5
            
        #    pc.get_all_neighbors_voronoi(real_atomdict, 0.0,
        #        system.triclinic, rot, rotinv,
        #        boxdims, 1)                
            
        #    pc.clean_voronoi_vertices(real_atomdict, 
        #        system.atoms, 0.0,
        #        system.triclinic, rot, rotinv,
        #        boxdims, cutoff)

            #unique_vertices = []
            #for i in range(len(system.vertex_is_unique)):
            #    for j in range(len(system.vertex_is_unique[i])):
            #        if system.vertex_is_unique[i][j]:
            #            unique_vertices.append(system.vertex_positions[i][j])

            #system.atoms["vertex_positions_unique_skipcheck"] = unique_vertices

        #assign extra options
        mapdict = {}
        mapdict["voronoi"] = {}
        mapdict["voronoi"]["volume"] = "voronoi_volume"
        mapdict["voronoi"]["face"] = {}
        mapdict["voronoi"]["face"]["vertices"] = "face_vertices"
        mapdict["voronoi"]["face"]["perimeters"] = "face_perimeters"
        mapdict["voronoi"]["vertex"] = {}
        mapdict["voronoi"]["vertex"]["vectors"] = "vertex_vectors"
        mapdict["voronoi"]["vertex"]["numbers"] = "vertex_numbers"
        mapdict["voronoi"]["vertex"]["positions"] = "vertex_positions"
        #mapdict["voronoi"]["vertex"]["unique_positions"] = "vertex_positions_unique_nofilter"
        system.atoms._add_attribute(mapdict)

    system.neighbors_found = True

def find_solids(system, bonds=0.5, threshold=0.5, avgthreshold=0.6, 
                      cluster=True, q=6, cutoff=0, right=True):
    """
    Distinguish solid and liquid atoms in the system.
    
    Parameters
    ----------
    bonds : int or float, optional
        Minimum number of solid bonds for an atom to be identified as
        a solid if the value is an integer. Minimum fraction of neighbors
        of an atom that should be solid for an atom to be solid if the
        value is float between 0-1. Default 0.5.
    
    threshold : double, optional
        Solid bond cutoff value. Default 0.5.
    
    avgthreshold : double, optional
        Value required for Averaged solid bond cutoff for an atom to be identified
        as solid. Default 0.6.
    
    cluster : bool, optional
        If True, cluster the solid atoms and return the number of atoms in the largest
        cluster.
    
    q : int, optional
        The Steinhardt parameter value over which the bonds have to be calculated.
        Default 6.
    
    cutoff : double, optional
        Separate value used for cluster classification. If not specified, cutoff used
        for finding neighbors is used.
    
    right: bool, optional
        If true, greater than comparison is to be used for finding solid particles. 
        default True.
    
    Returns
    -------
    solid : int
        Size of the largest solid cluster. Returned only if `cluster=True`.
    
    Notes
    -----
    The neighbors should be calculated before running this function.
    Check :func:`~pyscal.core.System.find_neighbors` method.
    
    `bonds` define the number of solid bonds of an atom to be identified as solid.
    Two particles are said to be 'bonded' if [1],
    .. math:: s_{ij} = \sum_{m=-6}^6 q_{6m}(i) q_{6m}^*(i) \geq threshold
    where `threshold` values is also an optional parameter.
    
    If the value of `bonds` is a fraction between 0 and 1, at least that much of an atom's neighbors
    should be solid for the atom to be solid.
    
    An additional parameter `avgthreshold` is an additional parameter to improve solid-liquid distinction.
    
    In addition to having a the specified number of `bonds`,
    
    .. math::  \langle s_{ij} \\rangle > avgthreshold
    
    also needs to be satisfied. In case another q value has to be used for calculation of S_ij, it can be
    set used the `q` attribute. In the above formulations, `>` comparison for `threshold` and `avgthreshold`
    can be changed to `<` by setting the keyword `right` to False.
    
    If `cluster` is True, a clustering is done for all solid particles. See :func:`~pyscal.csystem.find_clusters`
    for more details. 
    
    References
    ----------
    .. [1] Auer, S, Frenkel, D. Adv Polym Sci 173, 2005
    """
    #check if neighbors are found
    system._check_neighbors()

    if not isinstance(q, int):
        raise TypeError("q should be interger value")

    if not isinstance(threshold, (int, float)):
        raise TypeError("threshold should be a float value")
    else:
        if not ((threshold >= 0 ) and (threshold <= 1 )):
            raise ValueError("Value of threshold should be between 0 and 1")

    if not isinstance(avgthreshold, (int, float)):
        raise TypeError("avgthreshold should be a float value")
    else:
        if not ((avgthreshold >= 0 ) and (avgthreshold <= 1 )):
            raise ValueError("Value of avgthreshold should be between 0 and 1")

    #start identification routine
    #check the value of bonds and set criteria depending on that
    if isinstance(bonds, int):
        criteria = 0
    elif isinstance(bonds, float):
        if ((bonds>=0) and (bonds<=1.0)):
            criteria = 1
        else:
            raise TypeError("bonds if float should have value between 0-1")
    else:
         raise TypeError("bonds should be interger/float value")

    if right:
        compare_criteria = 0
    else:
        compare_criteria = 1

    system.calculate.steinhardt_parameter(q)

    #calculate solid neighs
    pc.calculate_bonds(system.atoms, q, 
        threshold, avgthreshold, bonds, 
        compare_criteria, criteria)

    mapdict = {}
    mapdict["solid"] = "solid"
    mapdict["steinhardt"] = {}
    mapdict["steinhardt"]["order"] = {}
    mapdict["steinhardt"]["order"]["bonds"] = "bonds"
    mapdict["steinhardt"]["order"]["sij"] = {}
    mapdict["steinhardt"]["order"]["sij"]["norm"] = "sij"
    mapdict["steinhardt"]["order"]["sij"]["average"] = "avg_sij"
    mapdict["steinhardt"]["order"]["sij"]["solid"] = "solid"
    system.atoms._add_attribute(mapdict)
    
    if cluster:
        lc = cluster_atoms(system, system.atoms.steinhardt.order.sij.solid, largest=True)
        return lc

def find_largest_cluster(system):
    """
    Find largest cluster among given clusters
    
    Parameters
    ----------
    None

    Returns
    -------
    lc : int
        Size of the largest cluster.
    """
    if not "cluster" in system.atoms.keys():
        raise RuntimeError("cluster_atoms needs to be called first")

    clusterlist = [x for x in system.atoms["cluster"] if x != -1]
    xx, xxcounts = np.unique(clusterlist, return_counts=True)
    arg = np.argsort(xxcounts)[-1]
    largest_cluster_size = xxcounts[arg]
    largest_cluster_id = xx[arg]

    system.atoms["largest_cluster"] = [True if system.atoms["cluster"][x]==largest_cluster_id else False for x in range(len(system.atoms["cluster"]))]
    
    mapdict = {}
    mapdict["cluster"] = {}
    mapdict["cluster"]["largest"] = "largest_cluster"
    system.atoms._add_attribute(mapdict)

    return largest_cluster_size


def cluster_atoms(system, condition, largest = True, cutoff=0):
    """
    Cluster atoms based on a property
    
    Parameters
    ----------
    condition : callable or atom property
        Either function which should take an :class:`~Atom` object, and give a True/False output
        or an attribute of atom class which has value or 1 or 0.
    
    largest : bool, optional
        If True returns the size of the largest cluster. Default False.
    
    cutoff : float, optional
        If specified, use this cutoff for calculation of clusters. By default uses the cutoff
        used for neighbor calculation.
    
    Returns
    -------
    lc : int
        Size of the largest cluster. Returned only if `largest` is True.
    
    Notes
    -----
    This function helps to cluster atoms based on a defined property. This property
    is defined by the user through the argument `condition` which is passed as a parameter.
    `condition` should be a boolean array the same length as number of atoms in the system.
    """
    
    system.apply_selection(condition=condition)
    pc.find_clusters(system.atoms, cutoff)

    mapdict = {}
    mapdict["cluster"] = {}
    mapdict["cluster"]["id"] = "cluster"
    system.atoms._add_attribute(mapdict)

    #done!
    lc = find_largest_cluster(system)
    #pcs.System.get_largest_cluster_atoms(system)
    system.remove_selection()
    if largest:
        return lc



import pytest
import numpy as np
import pyscal3.core as pc

def test_ordered_disorder():
    sys = pc.System('tests/files/conf.fcc.dump')
    sys.find.neighbors(method='cutoff', cutoff=0)
    sys.calculate.steinhardt_parameter(6)
    sys.calculate.disorder(averaged=True)
    assert np.mean(sys.atoms.steinhardt.disorder.norm) < 0.50
    assert np.mean(sys.atoms.steinhardt.disorder.average) < 0.50
    
def test_disordered_disorder():
    sys = pc.System('tests/files/conf.lqd.dump')
    sys.find.neighbors(method='cutoff', cutoff=0)
    sys.calculate.steinhardt_parameter(6)
    sys.calculate.disorder(averaged=True)
    assert np.mean(sys.atoms.steinhardt.disorder.norm) > 1.00
    assert np.mean(sys.atoms.steinhardt.disorder.average) > 1.00



"""


"""


import os
import numpy as np
import warnings
import itertools
from ase.io import write
import uuid
import gzip
import io
from scipy.special import sph_harm
import copy
from functools import partial, update_wrapper

from pyscal3.attributes import read_yaml
from pyscal3.atoms import Atoms, AttrSetter
import pyscal3.csystem as pc
import pyscal3.traj_process as ptp
from pyscal3.formats.ase import convert_snap
import pyscal3.structure_creator as pcs
from pyscal3.grain_boundary import GrainBoundary

import pyscal3.operations.operations as operations
import pyscal3.operations.cna as cna
import pyscal3.operations.centrosymmetry as centrosymmetry
import pyscal3.operations.neighbor as neighbor
import pyscal3.operations.input as inputmethods
import pyscal3.operations.calculations as calculations
import pyscal3.operations.identify as identify
import pyscal3.operations.voronoi as voronoi
import pyscal3.operations.chemical as chemical
import pyscal3.operations.visualize as visualize

#import pyscal.routines as routines
#import pyscal.visualization as pv

structure_dict = read_yaml(os.path.join(os.path.dirname(__file__), "data/structure_data.yaml"))
element_dict = read_yaml(os.path.join(os.path.dirname(__file__), "data/element_data.yaml"))

def _make_crystal(structure, 
    lattice_constant = 1.00, 
    repetitions = None, 
    ca_ratio = 1.633, 
    noise = 0, 
    element=None,
    primitive=False):
    
    atoms, box, sdict = pcs.make_crystal(structure, 
        lattice_constant=lattice_constant,
        repetitions=repetitions, 
        ca_ratio=ca_ratio,
        noise=noise, 
        element=element, 
        return_structure_dict=True,
        primitive=primitive)
    
    s = System()
    s.box = box
    s.atoms = atoms
    s.atoms._lattice = structure
    s.atoms._lattice_constant = lattice_constant
    s._structure_dict = sdict
    return s

def _make_general_lattice(positions,
    types, 
    box,
    lattice_constant = 1.00, 
    repetitions = None, 
    noise = 0,
    element=None):

    atoms, box, sdict = pcs.general_lattice(positions,
        types,
        box,
        lattice_constant=lattice_constant,
        repetitions=repetitions,
        noise=noise,
        element=element,
        return_structure_dict=True)
    s = System()
    s.box = box
    s.atoms = atoms
    s.atoms._lattice = 'custom'
    s.atoms._lattice_constant = lattice_constant
    s._structure_dict = sdict
    return s

def _make_grain_boundary(axis, 
    sigma, gb_plane,
    structure = None,
    element = None, 
    lattice_constant = 1,
    repetitions = (1,1,1),
    overlap=0.0):

    gb = GrainBoundary()
    gb.create_grain_boundary(axis=axis, sigma=sigma, 
                             gb_plane=gb_plane)

    if structure is not None:
        atoms, box, sdict = gb.populate_grain_boundary(structure, 
                                        repetitions = repetitions,
                                        lattice_parameter = lattice_constant,
                                        overlap=overlap)
    elif element is not None:
        atoms, box, sdict = gb.populate_grain_boundary(element, 
                                        repetitions=repetitions,
                                        overlap=overlap)
    s = System()
    s.box = box
    s.atoms = atoms
    s.atoms._lattice = structure
    s.atoms._lattice_constant = lattice_constant
    s._structure_dict = sdict
    #s = operations.remap_to_box(s)
    return s


class System:
    """
    Python class for holding the properties of an atomic configuration 
    """ 
    #system wide things available before structure creation
    create = AttrSetter()
    #create.head = pcs
    mapdict = {}
    mapdict["lattice"] = {}
    for key in structure_dict.keys():
        mapdict["lattice"][key] = update_wrapper(partial(_make_crystal, key), 
            _make_crystal)
    mapdict["lattice"]["custom"] = _make_general_lattice

    mapdict["element"] = {}
    for key in element_dict.keys():
        mapdict["element"][key] = update_wrapper(partial(_make_crystal,
            element_dict[key]['structure'],
            lattice_constant=element_dict[key]['lattice_constant'],
            element = key), pcs.make_crystal)

    mapdict["defect"] = {}
    mapdict["defect"]["grain_boundary"] = _make_grain_boundary

    create._add_attribute(mapdict)

    def __init__(self, filename=None, format="lammps-dump", 
                                            compressed = False, customkeys=None):
        self.initialized = True
        self.neighbors_found = False
        self.neighbor_method = None
        self.ghosts_created = False
        self.actual_box = None

        #box parameters
        self.rot = [[0,0,0], [0,0,0], [0,0,0]]
        self.rotinv = [[0,0,0], [0,0,0], [0,0,0]]
        self.boxdims = [0,0,0]
        self.triclinic = 0
        self._structure_dict = None
        self._atoms = Atoms()
        
        if filename is not None:
            inputmethods.read_inputfile(self, filename, 
                format=format, 
                compressed = compressed, 
                customkeys=customkeys)

        #customised methods for the class
        self.modify = AttrSetter()
        self.modify.head = operations
        mapdict = {}
        #repeat methid
        mapdict["repeat"] = update_wrapper(partial(operations.repeat, self), operations.repeat)
        mapdict["transform_to_cubic_cell"] = update_wrapper(partial(operations.extract_cubic_representation, self), operations.extract_cubic_representation)
        mapdict["remap_to_box"] = update_wrapper(partial(operations.remap_to_box, self), operations.remap_to_box)
        mapdict["embed_in_cubic_box"] = update_wrapper(partial(operations.embed_in_cubic_box, self), operations.embed_in_cubic_box)
        self.modify._add_attribute(mapdict)

        self.find = AttrSetter()
        mapdict = {}
        mapdict['neighbors'] = update_wrapper(partial(identify.find_neighbors, self), identify.find_neighbors)
        mapdict['solids'] = update_wrapper(partial(identify.find_solids, self), identify.find_solids)
        mapdict['clusters'] = update_wrapper(partial(identify.cluster_atoms, self), identify.cluster_atoms)
        mapdict['largest_cluster'] = update_wrapper(partial(identify.find_largest_cluster, self), identify.find_largest_cluster)
        self.find._add_attribute(mapdict)

        self.calculate = AttrSetter()
        mapdict = {}
        mapdict['distance'] = update_wrapper(partial(neighbor.get_distance, self), neighbor.get_distance)
        mapdict['average_over_neighbors'] = update_wrapper(partial(calculations.average_over_neighbors, self), calculations.average_over_neighbors)
        mapdict['steinhardt_parameter'] = update_wrapper(partial(calculations.calculate_q, self), calculations.calculate_q)
        mapdict['disorder'] = update_wrapper(partial(calculations.calculate_disorder, self), calculations.calculate_disorder)        
        mapdict['radial_distribution_function'] = update_wrapper(partial(calculations.calculate_rdf, self), calculations.calculate_rdf)        
        mapdict['angular_criteria'] = update_wrapper(partial(calculations.calculate_angularcriteria, self), calculations.calculate_angularcriteria)        
        mapdict['chi_params'] = update_wrapper(partial(calculations.calculate_chiparams, self), calculations.calculate_chiparams)
        mapdict['common_neighbor_analysis'] = update_wrapper(partial(cna.calculate_cna, self), cna.calculate_cna)
        mapdict['diamond_structure'] = update_wrapper(partial(cna.identify_diamond, self), cna.identify_diamond)
        mapdict['centrosymmetry'] = update_wrapper(partial(centrosymmetry.calculate_centrosymmetry, self), centrosymmetry.calculate_centrosymmetry)
        mapdict['voronoi_vector'] = update_wrapper(partial(voronoi.calculate_vorovector, self), voronoi.calculate_vorovector)
        self.calculate._add_attribute(mapdict)

        self.analyze = AttrSetter()
        mapdict = {}
        mapdict['common_neighbor_analysis'] = update_wrapper(partial(cna.calculate_cna, self), cna.calculate_cna)
        mapdict['diamond_structure'] = update_wrapper(partial(cna.identify_diamond, self), cna.identify_diamond)
        mapdict['short_range_order'] = update_wrapper(partial(chemical.calculate_sro, self), chemical.calculate_sro)
        self.analyze._add_attribute(mapdict)

        self.chemical = AttrSetter()
        mapdict = {}
        mapdict['short_range_order'] = update_wrapper(partial(chemical.calculate_sro, self), chemical.calculate_sro)
        self.chemical._add_attribute(mapdict)

        self.read = AttrSetter()
        mapdict = {}
        mapdict['file'] = update_wrapper(partial(inputmethods.read_inputfile, self), inputmethods.read_inputfile)
        mapdict['ase'] = update_wrapper(partial(inputmethods.read_inputfile, self, format='ase'), inputmethods.read_inputfile)
        self.read._add_attribute(mapdict)

        self.convert_to = AttrSetter()
        mapdict = {}
        mapdict['ase'] = update_wrapper(partial(convert_snap, self), convert_snap)
        self.convert_to._add_attribute(mapdict)

        self.write = AttrSetter()
        mapdict = {}
        mapdict['ase'] = update_wrapper(partial(convert_snap, self), convert_snap)
        mapdict['file'] = update_wrapper(partial(inputmethods.to_file, self), inputmethods.to_file)
        self.write._add_attribute(mapdict)

        self.show = AttrSetter()
        mapdict = {}
        mapdict['all'] = update_wrapper(partial(visualize.plot_simple, self), visualize.plot_simple)
        mapdict['continuous_property'] = update_wrapper(partial(visualize.plot_by_property, self), visualize.plot_by_property)
        mapdict['boolean_property'] = update_wrapper(partial(visualize.plot_by_boolean, self), visualize.plot_by_boolean)
        mapdict['selection'] = update_wrapper(partial(visualize.plot_by_selection, self), visualize.plot_by_selection)
        self.show._add_attribute(mapdict)


    def iter_atoms(self):
        return self.atoms.iter_atoms()

    @property
    def natoms(self):
        return self.atoms.natoms

    @property
    def concentration(self):
        return self.atoms.composition

    @property
    def composition(self):
        return self.atoms.composition

    @property
    def box(self):
        """
        Wrap for inbuilt box
        """
        if self.actual_box is not None:
            return self.actual_box
        else:
            return self._box

    @box.setter
    def box(self, userbox):
        """
        Box setter 
        """
        #we should automatically check for triclinic cells here
        self.internal_box = userbox
        self.actual_box = userbox

    @property
    def internal_box(self):
        return self._box

    @internal_box.setter
    def internal_box(self, userbox):
        """
        Box setter 
        """
        #we should automatically check for triclinic cells here
        summ = 0
        for i in range(3):
            box1 = np.array(userbox[i-1])
            box2 = np.array(userbox[i])
            summ += np.dot(box1, box2)/(np.linalg.norm(box1)*np.linalg.norm(box2))

        #check if the summ is zero
        if (np.abs(summ) > 1E-6):
            #this is a triclinic box
            rot = np.array(userbox).T
            rotinv = np.linalg.inv(rot)
            self.triclinic = 1
            self.rot = rot
            self.rotinv = rotinv

        self.boxdims[0] = np.sum(np.array(userbox[0])**2)**0.5
        self.boxdims[1] = np.sum(np.array(userbox[1])**2)**0.5
        self.boxdims[2] = np.sum(np.array(userbox[2])**2)**0.5

        #and we reset the original memory of the box
        self._box = userbox

    @property
    def box_dimensions(self):
        return [np.linalg.norm(self.box[x]) for x in range(3)]

    @property
    def direct_coordinates(self):
        dims = self.box_dimensions
        xfrac = np.array(self.atoms.positions)[:,0]/dims[0]
        yfrac = np.array(self.atoms.positions)[:,1]/dims[1]
        zfrac = np.array(self.atoms.positions)[:,2]/dims[2]
        coords = np.column_stack((xfrac, yfrac, zfrac))
        return coords    
    
    @property
    def volume(self):
        """
        Volume of box
        """
        vol = np.dot(np.cross(self._box[0], self._box[1]), self._box[2])
        return vol

    @property
    def atoms(self):
        return self._atoms
    
    @atoms.setter
    def atoms(self, atoms):
        """
        Set atoms
        """

        if(len(atoms['positions']) < 200):
            #we need to estimate a rough idea
            needed_atoms = 200 - len(atoms['positions'])
            #get a rough cell
            #print(needed_atoms)
            needed_cells = np.ceil(needed_atoms/len(atoms['positions']))
            nx = int(needed_cells**(1/3))
            if nx < 2:
                nx = 2
            if np.sum(self.box) == 0:
                raise ValueError("Simulation box should be initialized before atoms")
            atoms, box = operations.repeat(self, (nx, nx, nx), atoms=atoms, ghost=True, return_atoms=True)
            self.actual_box = self.box.copy()
            self.internal_box = box

        self._atoms = atoms
        


    def add_atoms(self, atoms):
        """
        Cleanly add a given list of atoms

        Parameters
        ----------
        atoms : dict

        Returns
        -------
        None
        """ 
        ## MOVE TO ATOMS
        self._atoms.add_atoms(atoms)
     
    def apply_selection(self, ids=None, indices=None, condition=None):
        self._atoms.apply_selection(ids=ids, indices=indices, condition=condition)    
    
    def remove_selection(self, ids=None, indices=None, condition=None):
        self._atoms.remove_selection(ids=ids, indices=indices, condition=condition)
    
    def delete(self, ids=None, indices=None, condition=None, selection=False):
        self._atoms.delete(ids=ids, indices=indices, condition=condition, selection=selection)



    def reset_neighbors(self):
        """
        Reset the neighbors of all atoms in the system.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        It is used automatically when neighbors are recalculated.

        """
        neighbor.reset_neighbors(self)


    def _check_neighbors(self):
        """
        Check if neighbors are calculated
        """
        if not self.neighbors_found:
            raise ValueError("This calculation needs neighbors to be calculated")


    


import pyscal3.csystem as pc

def calculate_centrosymmetry(system, nmax=12):
    """
    Calculate the centrosymmetry parameter

    Parameters
    ----------
    system: System object

    nmax : int, optional
        number of neighbors to be considered for centrosymmetry 
        parameters. Has to be a positive, even integer. Default 12

    Returns
    -------
    None
    
    Notes
    -----
    Calculate the centrosymmetry parameter for each atom which can be accessed by
    :attr:`~pyscal.catom.centrosymmetry` attribute. It calculates the degree of inversion
    symmetry of an atomic environment. Centrosymmetry recalculates the neighbor using
    the number method as specified in :func:`¬pyscal.core.System.find_neighbors` method. This
    is the ensure that the required number of neighbors are found for calculation of the parameter.

    The Greedy Edge Selection (GES) [1] as specified in [2] in used in this method. 
    GES algorithm is implemented in LAMMPS and Ovito. Please see [2] for
    a detailed description of the algorithms. 
    References
    ----------
    .. [1] Stukowski, A, Model Simul Mater SC 20, 2012
    .. [2] Larsen, arXiv:2003.08879v1, 2020

    """
    if not nmax>0:
        raise ValueError("nmax cannot be negative")

    if not nmax%2 == 0:
        raise ValueError("nmax has to even integer")

    system.atoms.create_attribute('centrosymmetry', fill_with = 0)
    system.find.neighbors(method='number', nmax=nmax, assign_neighbor=True)
    pc.calculate_centrosymmetry(system.atoms, nmax)

    return system.atoms.centrosymmetry



from pyscal3.operations.symmetry import get_symmetry
from pyscal3.core import System

def test_bcc():
	sys = System.create.element.Fe()
	sym = get_symmetry(sys)
	assert sym["international_symbol"] == 'Im-3m'	

def test_fcc():
	sys = System.create.element.Cu()
	sym = get_symmetry(sys)
	assert sym["international_symbol"] == 'Fm-3m'	

def test_hcp():
	sys = System.create.element.Mg()
	sym = get_symmetry(sys)
	assert sym["international_symbol"] == 'P6_3/mmc'	

def test_diamond():
	sys = System.create.element.Si()
	sym = get_symmetry(sys)
	assert sym["international_symbol"] == 'Fd-3m'	


import numpy as np
import itertools
import pyscal3.csystem as pc

def average_over_neighbors(system, key, include_system=True):
    """
    Perform a simple average over neighbor atoms

    Parameters
    ----------
    key: string
        atom property

    include_system: bool, optional
        If True, include the host atom in the calculation

    Returns
    -------

    """
    system._check_neighbors()

    if not key in system.atoms.keys():
        raise KeyError("required property not found!")

    test = system.atoms[key][0]

    if isinstance(test, list):
        raise TypeError("Averaging can only be done over 1D quantities")

    avgarr = []
    for i in range(len(system.atoms["positions"])):
        arr = []
        if include_system:
            arr.append(system.atoms[key][i])
        for j in system.atoms["neighbors"][i]:
            arr.append(system.atoms[key][j])
        avgarr.append(np.mean(arr))
    
    return avgarr 

def calculate_q(system, q, averaged=False, continuous_algorithm=False):
    """
    Find the Steinhardt parameter q_l for all atoms.

    Parameters
    ----------
    q : int or list of ints
        A list of all Steinhardt parameters to be found.

    averaged : bool, optional
        If True, return the averaged q values, default False
    
    continuous_algorithm: bool, optional
        See Notes for description.

    Returns
    -------
    q : list of floats
        calculated q values

    Notes
    -----
    Enables calculation of the Steinhardt parameters [1] q. The type of
    q values depend on the method used to calculate neighbors. See the description
    :func:`~pyscal.core.System.find_neighbors` for more details. 

    The option `continuous_algorithm` specifies which algorithm to use for calculations. If False, 
    an algorithm [3] is used. The C++ algorithm is faster is a large, consecutive number of q values (> 200)
    are to be calculated.

    This function creates three new attributes for this class: `qx`, `qx_real` and `qx_imag`,
    where `stands` for the q number.   

    References
    ----------
    .. [1] Steinhardt, PJ, Nelson, DR, Ronchetti, M. Phys Rev B 28, 1983
    .. [2] Lechner, W, Dellago, C, J Chem Phys, 2013
    """
    if isinstance(q, int):
        qq = [q]
    else:
        qq = q

    system._check_neighbors()

    if averaged:
        _calculate_aq(system, qq)
        qvals = [system.atoms["avg_q%d"%x] for x in qq]
    else:    
        if continuous_algorithm:
            lm = max(qq)
            pc.calculate_q(system.atoms, lm)
        else:
            _calculate_q(system, qq)
        qvals = [system.atoms["q%d"%x] for x in qq]
    return qvals

def _calculate_q(system, qq):
    """
    Private method for calculation of qvals
    """
    for val in qq:
        pc.calculate_q_single(system.atoms, val)

    mapdict = {}
    mapdict["steinhardt"] = {}
    mapdict["steinhardt"]["generic"] = {}
    for val in qq:
        key1a = "q%d_norm"%val
        key1b = "q%d"%val
        key2 = "q%d_real"%val
        key3 = "q%d_imag"%val
        mapdict["steinhardt"]["generic"][key1a] = key1b
        mapdict["steinhardt"]["generic"][key2] = key2
        mapdict["steinhardt"]["generic"][key3] = key3
    system.atoms._add_attribute(mapdict)


def _calculate_aq(system, qq):
    """
    Private method for calculation of avged qvals
    """

    todo_q = []
    for q in qq:
        keys = ["q%d"%q, "q%d_real"%q, "q%d_imag"%q]
        prod = []
        for key in keys:
            if key in system.atoms.keys():
                prod.append(True)
            else:
                prod.append(False)
        prod = np.prod(prod)
        if not prod:
            todo_q.append(q)

    _ = _calculate_q(system, todo_q)

    #loop over atoms
    for val in qq:
        pc.calculate_aq_single(system.atoms, val)

    mapdict = {}
    mapdict["steinhardt"] = {}
    mapdict["steinhardt"]["average"] = {}
    for val in qq:
        key1a = "q%d_norm"%val
        key1b = "q%d"%val
        key2 = "q%d_real"%val
        key3 = "q%d_imag"%val
        mapdict["steinhardt"]["average"][key1a] = key1b
        mapdict["steinhardt"]["average"][key2] = key2
        mapdict["steinhardt"]["average"][key3] = key3
    system.atoms._add_attribute(mapdict)

def calculate_disorder(system, averaged=False, q=6):
    """
    Calculate the disorder criteria for each atom
    
    Parameters
    ----------
    averaged : bool, optional
        If True, calculate the averaged disorder. Default False.
    q : int, optional
        The Steinhardt parameter value over which the bonds have to be calculated.
        Default 6.
    
    Returns
    -------
    None
    
    Notes
    -----
    Calculate the disorder criteria as introduced in [1]. The disorder criteria value for each atom is defined by,
    .. math::
        D_j = \\frac{1}{N_b^j} \sum_{i=1}^{N_b} [ S_{jj} + S_{kk} -2S_{jk}]
    where .. math:: S_{ij} = \sum_{m=-6}^6 q_{6m}(i) q_{6m}^*(i)
    
    Any q value other than six can also be used. This can be specified using the `q` argument.

    The keyword `averaged` is True, the disorder value is averaged over the atom and its neighbors. 
    For ordered systems, the value of disorder would be zero which would increase
    and reach one for disordered systems.

    This function creates two new attributes for this class: `disorder` and `avg_disorder`.
    
    References
    ----------
    .. [1] Kawasaki, T, Onuki, A, J. Chem. Phys. 135, 2011
    """
    #now routine for calculation of disorder

    keys = ["q%d_real"%q, "q%d_imag"%q]
    prod = []
    for key in keys:
        if key in system.atoms.keys():
            prod.append(True)
        else:
            prod.append(False)
    prod = np.prod(prod)
    if not prod:
        calculate_q(system, q)

    pc.calculate_disorder(system.atoms, q)

    mapdict = {}
    mapdict["steinhardt"] = {}
    mapdict["steinhardt"]["disorder"] = {}
    mapdict["steinhardt"]["disorder"]["norm"] = "disorder"

    if averaged:
        #average the disorder
        avg_arr = system.calculate.average_over_neighbors("disorder")
        system.atoms["avg_disorder"] = avg_arr
        mapdict["steinhardt"]["disorder"]["average"] = "avg_disorder"
    system.atoms._add_attribute(mapdict)

def calculate_rdf(system, rmin=0, rmax=5.00, bins=100):
    """
    Calculate the radial distribution function.
    
    Parameters
    ----------
    rmin : float, optional
        minimum value of the distance histogram. Default 0.0.
    
    rmax : float, optional
        maximum value of the distance histogram. Default 5.0

    bins : int
        number of bins in the histogram
            
    Returns
    -------
    rdf : array of ints
        Radial distribution function
    
    r : array of floats
        radius in distance units
    """
    system.find.neighbors(method="cutoff", cutoff=rmax)
    distances = list(itertools.chain(*system.atoms["neighbordist"]))

    hist, bin_edges = np.histogram(distances, bins=bins, 
        range=(rmin, rmax), density=True)
    edgewidth = np.abs(bin_edges[1]-bin_edges[0])
    hist = hist.astype(float)
    r = bin_edges[:-1]

    #get box density
    rho = system.natoms/system.volume
    shell_vols = (4./3.)*np.pi*((r+edgewidth)**3 - r**3)
    shell_rho = hist/shell_vols
    #now divide to get final value
    rdf = shell_rho/rho
    return rdf, r

def calculate_angularcriteria(system):
    """
    Calculate the angular criteria for each atom
    
    Parameters
    ----------
    None
    
    Returns
    -------
    None
    
    Notes
    -----
    Calculates the angular criteria for each atom as defined in [1]_. Angular criteria is
    useful for identification of diamond cubic structures. Angular criteria is defined by,
    
    .. math::
        A = \sum_{i=1}^6 (\cos(\\theta_i) + \\frac{1}{3})^2
    
    where cos(theta) is the angle size suspended by each pair of neighbors of the central
    atom. A will have a value close to 0 for structures if the angles are close to 109 degrees.
    The calculated A parameter for each atom can be accessed by system.angular
    
    References
    ----------
    .. [1] Uttormark, MJ, Thompson, MO, Clancy, P, Phys. Rev. B 47, 1993
    """
    system._check_neighbors()
    angulars = []

    for count, pos1 in enumerate(system.atoms["positions"]):
        
        dists = []
        distneighs = []
        distvectors = []

        for count2, neigh in enumerate(system.atoms["neighbors"][count]):
            pos2 = system.atoms["positions"][neigh]
            dist = system.atoms["neighbordist"][count][count2]
            vectors = system.atoms["diff"][count][count2]
            dists.append(dist)
            distneighs.append(neigh)
            distvectors.append(vectors)

        args = np.argsort(dists)
        #find top four
        topfourargs = np.array(args)[:4]

        combos = list(itertools.combinations(topfourargs, 2))
        costhetasum = 0

        for combo in combos:
            vec1 = distvectors[combo[0]]
            vec2 = distvectors[combo[1]]
            modvec1 = np.sqrt(np.sum([x**2 for x in vec1]))
            modvec2 = np.sqrt(np.sum([x**2 for x in vec2]))
            costheta = np.dot(vec1, vec2)/(modvec1*modvec2)
            costhetasum += (costheta +(1./3.))**2
        angulars.append(costhetasum)

    system.atoms["angular"] = angulars
    
    mapdict = {}
    mapdict["angular_parameters"] = {}
    mapdict["angular_parameters"]["diamond_angle"] = "angular"
    system.atoms._add_attribute(mapdict)

def calculate_chiparams(system, angles=False):
    """
    Calculate the chi param vector for each atom
    
    Parameters
    ----------
    angles : bool, optional
        If True, return the list of cosines of all neighbor pairs
    
    Returns
    -------
    angles : array of floats
        list of all cosine values, returned only if `angles` is True.
    
    Notes
    -----
    This method tries to distinguish between crystal structures by finding the cosines of angles
    formed by an atom with its neighbors. These cosines are then historgrammed with bins
    `[-1.0, -0.945, -0.915, -0.755, -0.705, -0.195, 0.195, 0.245, 0.795, 1.0]` to find a vector for
    each atom that is indicative of its local coordination. Compared to chi parameters from chi_0 to
    chi_7 in the associated publication, the vector here is from chi_0 to chi_8. This is due to an additional
    chi parameter which measures the number of neighbors between cosines -0.705 to -0.195.
    Parameter `nlimit` specifies the number of nearest neighbors to be included in the analysis to find the cutoff.
    If parameter `angles` is true, an array of all cosine values is returned. The publication further provides
    combinations of chi parameters for structural identification which is not implemented here. The calculated
    chi params can be accessed using :attr:`~pyscal.catom.chiparams`.
    
    References
    ----------
    .. [1] Ackland, Jones, Phys. Rev. B 73, 2006
    """

    system._check_neighbors()

    bins = [-1.0, -0.945, -0.915, -0.755, -0.705, -0.195, 0.195, 0.245, 0.795, 1.0]
    chiparams = []
    cosines = []

    for count, pos in enumerate(system.atoms["positions"]):

        dists = system.atoms["neighbordist"][count]
        neighs = system.atoms["neighbors"][count]

        args = range(len(dists))
        combos = list(itertools.combinations(args, 2))
        costhetas = []
        
        for combo in combos:
            vec1 = system.atoms["diff"][count][combo[0]]
            vec2 = system.atoms["diff"][count][combo[1]]
            modvec1 = np.linalg.norm(vec1)
            modvec2 = np.linalg.norm(vec2)
            costheta = np.dot(vec1, vec2)/(modvec1*modvec2)
            #found costheta
            costhetas.append(costheta)


        #now add according to classification in paper
        chivector = np.histogram(costhetas, bins=bins)
        chiparams.append(chivector[0])
        if angles:
            cosines.append(costhetas)
    
    system.atoms["chiparams"] = chiparams
    
    mapdict = {}
    mapdict["angular_parameters"] = {}
    mapdict["angular_parameters"]["chi_params"] = "chiparams"
    
    if angles:
        system.atoms["cosines"] = cosines
        mapdict["angular_parameters"]["cosines"] = "cosines"

    system.atoms._add_attribute(mapdict)



from pyscal3.core import System
import numpy as np


def test_sro():
	sys = System.create.lattice.l12(lattice_constant=4, repetitions=(2,2,2))
	sys.find.neighbors(method='cutoff')
	sys.chemical.short_range_order()
	assert sys.atoms.chemical.short_range_order[0] == 1
	assert np.isclose(sys.atoms.chemical.short_range_order[1], -0.33333333)


import pytest
import os,sys,inspect
import numpy as np
import pyscal3.core as pc

def test_angular():
    sys = pc.System.create.lattice.diamond(repetitions = [4, 4, 4])
    sys.find.neighbors(method = 'cutoff', cutoff=0)
    sys.calculate.angular_criteria()

    assert np.round(np.mean(np.array(sys.atoms.angular_parameters.diamond_angle)), decimals=2) == 0.00




import numpy as np
import warnings
import os
import copy

from pyscal3.atoms import Atoms
from pyscal3.attributes import read_yaml
structures = read_yaml(os.path.join(os.path.dirname(__file__), "data/structure_data.yaml"))

def make_crystal(structure, 
    lattice_constant = 1.00, 
    repetitions = None, 
    ca_ratio = 1.633, 
    noise = 0, 
    element=None,
    return_structure_dict=False, 
    structures=structures,
    primitive=False):
    """
    Create a basic crystal structure and return it as a list of `Atom` objects
    and box dimensions.

    Parameters
    ----------
    structure : {'sc', 'bcc', 'fcc', 'hcp', 'diamond', 'a15' or 'l12'}
        type of the crystal structure

    lattice_constant : float, optional
        lattice constant of the crystal structure, default 1

    repetitions : list of ints of len 3, optional
        of type `[nx, ny, nz]`, repetions of the unit cell in x, y and z directions.
        default `[1, 1, 1]`.

    ca_ratio : float, optional
        ratio of c/a for hcp structures, default 1.633

    noise : float, optional
        If provided add normally distributed noise with standard deviation `noise` to the atomic positions.

    Returns
    -------
    atoms : list of `Atom` objects
        list of all atoms as created by user input

    box : list of list of floats
        list of the type `[[xlow, xhigh], [ylow, yhigh], [zlow, zhigh]]` where each of them are the lower
        and upper limits of the simulation box in x, y and z directions respectively.

    Examples
    --------
    >>> atoms, box = make_crystal('bcc', lattice_constant=3.48, repetitions=[2,2,2])
    >>> sys = System()
    >>> sys.assign_atoms(atoms, box)

    """
    if repetitions == None:
        repetitions = [1, 1, 1]
    elif isinstance(repetitions, int):
        repetitions = [repetitions, repetitions, repetitions]

    if structure in structures.keys():
        if primitive:
            if 'primitive' not in structures[structure].keys():
                raise ValueError('primitive not found, try setting primitive=False')
            sdict = copy.copy(structures[structure]['primitive']) 
        else:
            if 'conventional' not in structures[structure].keys():
                raise ValueError('conventional not found, try setting primitive=True')
            sdict = copy.copy(structures[structure]['conventional']) 
    else:
        raise ValueError("Unknown crystal structure")
    
    unique_types = np.unique(sdict["species"])
    #print(element)
    if element is not None:
        if isinstance(element, str):
            element = [element]
        if not (len(element) == len(unique_types)):
            raise ValueError("Elements should equal number of species in the system")
        
        element_dict = dict([x for x in zip(unique_types, element)])
    else:
        element = [None for x in range(len(unique_types))]
    element_dict = dict([x for x in zip(unique_types, element)])
    

    #fix ca ration
    if structure in ['hcp', 'dhcp']:
        sdict['box'][2][2] = ca_ratio 

    #from here, the creation starts
    box = lattice_constant*np.array(sdict["box"])        

    pos = np.array([_unfold_positions(p, box) for p in sdict["positions"]])
    pos = np.array([_generate_noise(x, noise) for x in pos])
    nop = len(pos)
    types = sdict['species']
    species = [element_dict[typ] for typ in types] 

    for d in range(3):        
        pos_list = []
        for i in range(1, repetitions[d]):
            npos = copy.copy(pos)
            npos = npos + i*box[d]
            pos_list.append([_generate_noise(n, noise) for n in npos])

        pos = np.concatenate((pos, *pos_list))
        types = np.concatenate((types, np.tile(types, len(pos_list))))
        species = np.concatenate((species, np.tile(species, len(pos_list))))

                    
    atoms = {}
    atoms['positions'] = pos
    atoms['ids'] = [x+1 for x in range(len(pos))]
    atoms['types'] = types
    atoms['species'] = species
    atoms['ghost'] = [False for x in range(len(types))]

    patoms = Atoms()
    patoms.from_dict(atoms)

    #scale box 
    box[0] = repetitions[0]*np.array(box[0])
    box[1] = repetitions[1]*np.array(box[1])
    box[2] = repetitions[2]*np.array(box[2])

    if return_structure_dict:
        return patoms, box, sdict

    return patoms, box

def general_lattice(positions,
    types, 
    box,
    lattice_constant = 1.00, 
    repetitions = None, 
    noise = 0,
    element=None,
    return_structure_dict=False):
    """
    Create a general lattice structure.

    species: list
        list of per-atom species

    positions:
        list of relative positions positions of reach atom (between 0-1)

    scaling_fractors:
        factors with which the unit cell should be scaled, for example hcp could
        have [1,1.73, 1.63]. Default [1,1,1]

    lattice_constant : float, optional
        lattice constant of the crystal structure, default 1

    repetitions : list of ints of len 3, optional
        of type `[nx, ny, nz]`, repetions of the unit cell in x, y and z directions.
        default `[1, 1, 1]`.

    noise : float, optional
        If provided add normally distributed noise with standard deviation `noise` to the atomic positions.
    
    element : string, optional
        The chemical element
    """
    if not (len(types) == len(positions)):
        raise ValueError("types and positions should have same length!")

    sdict = {"custom":
                {"conventional":
                    {
                     "species": types,
                     "positions": positions,
                     "box": box,
                    }
                }
            }

    atoms, box = make_crystal("custom", lattice_constant=lattice_constant,
        repetitions=repetitions, noise=noise, element=element,
        structures=sdict)

    if return_structure_dict:
        return atoms, box, sdict

    return atoms, box


def _generate_noise(p, noise):
    if noise > 0:
        p[0] = np.random.normal(loc=p[0], scale=noise)
        p[1] = np.random.normal(loc=p[1], scale=noise)
        p[2] = np.random.normal(loc=p[2], scale=noise)
    return p

def _update_list_of_elements():
    """
    Only run when needed to update database
    """
    import mendeleev
    from mendeleev import element
    el_list = dir(mendeleev)
    el_dict = {}

    for el in el_list:
        if len(el) == 2:
            if el=="db":
                break
            chem = element(el)
            struct = chem.lattice_structure
            lc = chem.lattice_constant
            if (struct is not None) and (lc is not None):
                if struct.lower() in ['sc', 'bcc', 'fcc', 'hex', 'dia']:
                    el_dict[el] = {}
                    if struct.lower() == "dia":
                        struct = "diamond"
                    elif struct.lower() == "hex":
                        struct = "hcp"
                    el_dict[el]["structure"] = str(struct.lower())
                    el_dict[el]["lattice_constant"] = float(lc)
    return el_dict

def _unfold_positions(p, box):
    return p[0]*box[0] + p[1]*box[1] + p[2]*box[2]



import pyscal3.csystem as pc

def calculate_cna(system, lattice_constant=None):
    """
    Calculate the Common Neighbor Analysis indices

    Parameters
    ----------
    system: System object

    lattice_constant : float, optional
        lattice constant to calculate CNA. If not specified,
        adaptive CNA will be used

    Returns
    -------
    resdict: dict
        dictionary containing the calculated results

    Notes
    -----
    Performs the common neighbor analysis [1][2] or the adaptive common neighbor
    analysis [2] and assigns a structure to each atom.
    
    If `lattice_constant` is specified, a convential common neighbor analysis is
    used. If `lattice_constant` is not specified, adaptive common neighbor analysis is used. 
    The assigned structures can be accessed by :attr:`~pyscal.catom.Atom.structure`.
    The values assigned for stucture are 0 Unknown, 1 fcc, 2 hcp, 3 bcc, 4 icosahedral.

    References
    ----------
    .. [1] Faken, Jonsson, CMS 2, 1994
    .. [2] Honeycutt, Andersen, JPC 91, 1987
    .. [3] Stukowski, A, Model Simul Mater SC 20, 2012

    """
    system.atoms.create_attribute('structure', fill_with = 0)
    #run to calculate temp neighbors
    system.find.neighbors(method='number', nmax=14, assign_neighbor=False)

    if lattice_constant is None:
        #we need adaptive calculation
        pc.get_acna_neighbors_cn12(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims)
        pc.identify_cn12(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims)

        pc.get_acna_neighbors_cn14(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims)
        pc.identify_cn14(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims)

    else:
        system.lattice_constant = lattice_constant
        pc.get_cna_neighbors(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims,
            lattice_constant, 1)
        pc.identify_cn12(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims)

        pc.get_cna_neighbors(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims,
            lattice_constant, 2)
        pc.identify_cn14(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims)


    res_dict = {
    "others": len([x for x in system.atoms.structure if x==0]),
    "fcc": len([x for x in system.atoms.structure if x==1]),
    "hcp": len([x for x in system.atoms.structure if x==2]),
    "bcc": len([x for x in system.atoms.structure if x==3]),
    "ico": len([x for x in system.atoms.structure if x==4]),
    }
    return res_dict

def identify_diamond(system):
    """
    Identify diamond structure

    Parameters
    ----------
    system: System object

    Returns
    -------
    diamondstructure : dict
        dict of structure signature

    Notes
    -----
    Identify diamond structure using the algorithm mentioned in [1]. It is an
    extended CNA method. The integers 5, 6, 7, 8, 9 and 10 are assigned to the
    structure variable of the atom. 5 stands for cubic diamond, 6 stands for first
    nearest neighbors of cubic diamond and 7 stands for second nearest neighbors
    of cubic diamond. 8 signifies hexagonal diamond, the first nearest neighbors
    are marked with 9 and second nearest neighbors with 10.

    References
    ----------
    .. [1] Maras et al, CPC 205, 2016
    """
    system.atoms.create_attribute('structure', fill_with = 0)
    #run to calculate temp neighbors
    system.find.neighbors(method='number', nmax=4, assign_neighbor=False)

    pc.identify_diamond_cna(system.atoms, system.triclinic, 
            system.rot, system.rotinv, system.boxdims)

    res_dict = {
    "others": len([x for x in system.atoms.structure if x==0]),
    "cubic diamond": len([x for x in system.atoms.structure if x==5]),
    "cubic diamond 1NN": len([x for x in system.atoms.structure if x==6]),
    "cubic diamond 2NN": len([x for x in system.atoms.structure if x==7]),
    "hex diamond": len([x for x in system.atoms.structure if x==8]),
    "hex diamond 1NN": len([x for x in system.atoms.structure if x==9]),
    "hex diamond 2NN": len([x for x in system.atoms.structure if x==10]),
    }
    return res_dict    